"""
Utility functions for pywbem end2end tests.
"""

from __future__ import absolute_import, print_function

from copy import deepcopy

from pywbem import CIMInstance, CIMInstanceName, \
    Error, CIMError, CIM_ERR_INVALID_CLASS, WBEMConnection


def latest_profile_inst(profile_insts):
    """
    Determine the instance with the latest version in a list of registered
    profile instances and return that instance.
    """
    profile_versions = [inst['RegisteredVersion'] for inst in profile_insts]
    latest_i = None
    latest_version = [0, 0, 0]
    for i, profile_version in enumerate(profile_versions):
        profile_version = [int(v) for v in profile_version.split('.')]
        if profile_version > latest_version:
            latest_version = profile_version
            latest_i = i
    return profile_insts[latest_i]


def path_equal(inst1_path, inst2_path):
    """
    Return whether two instance paths are equal, with special treatment
    of their host component:
    - The host component is compared literally (but case insensitively),
      i.e. no IP address to hostname translation is performed.
    - If one or both of the instance paths have their host component set to
      None, the comparison is considered equal.

    For keys of the instance paths that are references, the special
    treatment of their host component is applied recursively.
    """
    assert isinstance(inst1_path, CIMInstanceName)
    assert isinstance(inst2_path, CIMInstanceName)
    if inst1_path.classname.lower() != inst2_path.classname.lower():
        return False
    if inst1_path.namespace.lower() != inst2_path.namespace.lower():
        return False
    if inst1_path.host is not None and \
            inst2_path.host is not None and \
            inst1_path.host.lower() != inst2_path.host.lower():
        return False
    if len(inst1_path.keybindings) != len(inst2_path.keybindings):
        return False
    for key_name in inst1_path.keybindings:
        key1_value = inst1_path.keybindings[key_name]
        if key_name not in inst2_path.keybindings:
            return False
        key2_value = inst2_path.keybindings[key_name]
        # pylint: disable=no-else-return
        if isinstance(key1_value, CIMInstanceName) and \
                isinstance(key2_value, CIMInstanceName) and \
                not path_equal(key1_value, key2_value):
            return False
        elif key1_value != key2_value:
            return False
    return True


def path_in(inst_path, inst_path_list):
    """
    Return whether an instance path is in an instance path list, with special
    treatment of their host component when testing for equality, as described
    for path_equal().

    Each item in inst_path list can be an instance path (CIMInstanceName)
    or an instance (CIMInstance) whose path is then used for the test.
    """
    assert isinstance(inst_path, CIMInstanceName)
    for ip in inst_path_list:
        if isinstance(ip, CIMInstance):
            ip = ip.path
        assert isinstance(ip, CIMInstanceName)
        if path_equal(inst_path, ip):
            return True
    return False


class ServerObjectCache(object):
    """
    A cache for named lists of CIM objects from a particular WBEM server.

    Each list is identified by a name that can be an arbitrary string.

    The WBEM server is identified by its URL. This allows the objects to
    be reused across WBEMConnection or WBEMServer objects to the same server.
    """

    def __init__(self):
        # Cache dictionary:
        # - key: server URL
        # - value: dictionary of named lists:
        #   - key: list name
        #   _ value: list of CIM objects (e.g. CIMInstanceName, CIMClass)
        self._server_dict = dict()

    def add_list(self, server_url, list_name, obj_list):
        """Add obj_list to the list named list_name"""
        if server_url not in self._server_dict:
            self._server_dict[server_url] = dict()
        list_dict = self._server_dict[server_url]
        if list_name in list_dict:
            raise KeyError(
                "List {0!r} for server {1!r} is already in the cache".
                format(list_name, server_url))
        list_dict[list_name] = deepcopy(obj_list)

    def del_list(self, server_url, list_name):
        """Delete list_name"""
        if server_url not in self._server_dict:
            raise KeyError(
                "Server {0!r} is not in the cache".
                format(server_url))
        list_dict = self._server_dict[server_url]
        if list_name not in list_dict:
            raise KeyError(
                "List {0!r} for server {1!r} is not in the cache".
                format(list_name, server_url))
        del list_dict[list_name]
        if not list_dict:
            del self._server_dict[server_url]

    def get_list(self, server_url, list_name):
        """Get the list named list_name"""
        if server_url not in self._server_dict:
            raise KeyError(
                "Server {0!r} is not in the cache".
                format(server_url))
        list_dict = self._server_dict[server_url]
        if list_name not in list_dict:
            raise KeyError(
                "List {0!r} for server {1!r} is not in the cache".
                format(list_name, server_url))
        return list_dict[list_name]

    def has_list(self, server_url, list_name):
        """Return boolean indicating whether there is a list named list_name"""
        if server_url not in self._server_dict:
            return False
        list_dict = self._server_dict[server_url]
        if list_name not in list_dict:
            return False
        return True


ENUM_INST_CACHE = ServerObjectCache()


def instance_of(conn, obj_list, classname):
    """
    Return whether all of a set of CIM instances and/or CIM instance paths are
    of a particular CIM class (including subclasses).

    Because there are WBEM servers without support for class operations,
    this is implemented without relying on class operations. The function
    performs an EnumerateInstanceNames operation on the desired class in
    the namespace of the instance in question, and verifies that the
    instance in question is in the result.

    Parameters:

        obj_list (CIMInstanceName or CIMInstance or tuple/list thereof):
          The CIM instances and CIM instance paths to be evaluated.

        classname (string): The CIM class name.
    """

    assert obj_list is not None
    if not isinstance(obj_list, (tuple, list)):
        obj_list = [obj_list]

    paths = []
    for obj in obj_list:
        if isinstance(obj, CIMInstance):
            path = obj.path
        else:
            path = obj
        assert isinstance(path, CIMInstanceName)  # ensured by CIMInstance
        assert path.classname is not None  # ensured by CIMInstanceName
        assert path.namespace is not None
        paths.append(path)

    assert len(paths) >= 1
    namespace = paths[0].namespace
    for path in paths:
        assert path.namespace.lower() == namespace.lower()

    enum_paths_id = namespace.lower() + ':' + classname.lower()

    try:
        enum_paths = ENUM_INST_CACHE.get_list(conn.url, enum_paths_id)
    except KeyError:
        try:
            enum_paths = conn.EnumerateInstanceNames(
                namespace=namespace, ClassName=classname,
                asserted=False)
        except CIMError as exc:
            if exc.status_code == CIM_ERR_INVALID_CLASS:
                raise AssertionError(
                    "Server {0} at {1}: Class {2!r} does not exist in "
                    "namespace {3!r}".
                    format(conn.server_definition.nickname, conn.url,
                           classname, namespace))
            conn.raise_as_assertion_error(exc, 'EnumerateInstanceNames')
        except Error as exc:
            conn.raise_as_assertion_error(exc, 'EnumerateInstanceNames')
        ENUM_INST_CACHE.add_list(conn.url, enum_paths_id, enum_paths)

    for path in paths:
        if not path_in(path, enum_paths):
            return False
    return True


class WBEMConnectionAsserted(WBEMConnection):
    """
    Subclass of WBEMConnection that adds the functionality to assert that
    the operation succeeds, and that raises an AssertionError if it does
    not succeed.

    The assertion behavior is disabled by default and can be enabled by
    specifying keyword argument `asserted=True`.
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters:

          server_definition(ServerDefinition): Server to test against
            (required keyword argument)

          All other positional and keyword arguments are passed to the
          superclass init.
        """
        server_definition = kwargs['server_definition']
        del kwargs['server_definition']
        super(WBEMConnectionAsserted, self).__init__(*args, **kwargs)
        self.server_definition = server_definition

    def _call_op(self, funcname, *args, **kwargs):
        """
        Call the specified operation function of the base class.

        If keyword argument 'asserted' is True, raise any pywbem.Error
        exceptions as AssertionError exceptions.
        """
        if 'asserted' in kwargs:
            asserted = bool(kwargs['asserted'])
            del kwargs['asserted']
        else:
            # The default must be non-asserted, in order to allow all the
            # operation invocations in WBEMServer to do their error handling.
            asserted = False
        func = getattr(super(WBEMConnectionAsserted, self), funcname)
        # returns a bound method
        if asserted:
            try:
                return func(*args, **kwargs)
            except Error as exc:
                self.raise_as_assertion_error(exc, funcname, *args, **kwargs)
        else:
            return func(*args, **kwargs)

    def raise_as_assertion_error(self, exc, funcname, *args, **kwargs):
        """
        Raise an AssertionError about the specified exception.
        """
        parm_list = ["{0!r}".format(a) for a in args]
        parm_list.extend(["{0}={1!r}".format(k, kwargs[k]) for k in kwargs])
        parm_str = ", ".join(parm_list)
        raise AssertionError(
            "Server {0} at {1}: WBEMConnection.{2}() failed and raised "
            "{3} - {4}. "
            "Parameters: ({5})".
            format(self.server_definition.nickname, self.url, funcname,
                   exc.__class__.__name__, exc, parm_str))

    def EnumerateInstances(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('EnumerateInstances', *args, **kwargs)

    def EnumerateInstanceNames(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('EnumerateInstanceNames', *args, **kwargs)

    def GetInstance(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('GetInstance', *args, **kwargs)

    def ModifyInstance(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('ModifyInstance', *args, **kwargs)

    def CreateInstance(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('CreateInstance', *args, **kwargs)

    def DeleteInstance(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('DeleteInstance', *args, **kwargs)

    def Associators(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('Associators', *args, **kwargs)

    def AssociatorNames(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('AssociatorNames', *args, **kwargs)

    def References(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('References', *args, **kwargs)

    def ReferenceNames(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('ReferenceNames', *args, **kwargs)

    def InvokeMethod(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('InvokeMethod', *args, **kwargs)

    def ExecQuery(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('ExecQuery', *args, **kwargs)

    def IterEnumerateInstances(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('IterEnumerateInstances', *args, **kwargs)

    def IterEnumerateInstancePaths(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('IterEnumerateInstancePaths', *args, **kwargs)

    def IterAssociatorInstances(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('IterAssociatorInstances', *args, **kwargs)

    def IterAssociatorInstancePaths(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('IterAssociatorInstancePaths', *args, **kwargs)

    def IterReferenceInstances(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('IterReferenceInstances', *args, **kwargs)

    def IterReferenceInstancePaths(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('IterReferenceInstancePaths', *args, **kwargs)

    def IterQueryInstances(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('IterQueryInstances', *args, **kwargs)

    def OpenEnumerateInstances(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('OpenEnumerateInstances', *args, **kwargs)

    def OpenEnumerateInstancePaths(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('OpenEnumerateInstancePaths', *args, **kwargs)

    def OpenAssociatorInstances(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('OpenAssociatorInstances', *args, **kwargs)

    def OpenAssociatorInstancePaths(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('OpenAssociatorInstancePaths', *args, **kwargs)

    def OpenReferenceInstances(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('OpenReferenceInstances', *args, **kwargs)

    def OpenReferenceInstancePaths(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('OpenReferenceInstancePaths', *args, **kwargs)

    def OpenQueryInstances(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('OpenQueryInstances', *args, **kwargs)

    def PullInstancesWithPath(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('PullInstancesWithPath', *args, **kwargs)

    def PullInstancePaths(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('PullInstancePaths', *args, **kwargs)

    def PullInstances(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('PullInstances', *args, **kwargs)

    def CloseEnumeration(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('CloseEnumeration', *args, **kwargs)

    def EnumerateClasses(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('EnumerateClasses', *args, **kwargs)

    def EnumerateClassNames(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('EnumerateClassNames', *args, **kwargs)

    def GetClass(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('GetClass', *args, **kwargs)

    def ModifyClass(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('ModifyClass', *args, **kwargs)

    def CreateClass(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('CreateClass', *args, **kwargs)

    def DeleteClass(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('DeleteClass', *args, **kwargs)

    def EnumerateQualifiers(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('EnumerateQualifiers', *args, **kwargs)

    def GetQualifier(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('GetQualifier', *args, **kwargs)

    def SetQualifier(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('SetQualifier', *args, **kwargs)

    def DeleteQualifier(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        return self._call_op('DeleteQualifier', *args, **kwargs)


def server_func_asserted(server, funcname, *args, **kwargs):
    """
    Call a method on a `WBEMServer` object and assert that it does not raise a
    pywbem.Error exception.
    """
    func = getattr(server, funcname)
    try:
        return func(*args, **kwargs)
    except Error as exc:
        parm_list = ["{0!r}".format(a) for a in args]
        parm_list.extend(["{0}={1!r}".format(k, kwargs[k]) for k in kwargs])
        parm_str = ", ".join(parm_list)
        raise AssertionError(
            "Server {0} at {1}: Calling WBEMServer.{2}() failed and "
            "raised {3} - {4}. "
            "Parameters: ({5})".
            format(server.conn.server_definition.nickname, server.conn.url,
                   funcname, exc.__class__.__name__, exc, parm_str))


def server_prop_asserted(server, propname):
    """
    Get the value of a property of a `WBEMServer` object and assert that it
    does not raise a pywbem.Error exception.
    """
    try:
        return getattr(server, propname)
    except Error as exc:
        raise AssertionError(
            "Server {0} at {1}: Getting WBEMServer.{2} failed and "
            "raised {3} - {4}.".
            format(server.conn.server_definition.nickname, server.conn.url,
                   propname, exc.__class__.__name__, exc))



#!/usr/bin/env python

"""
Test _listener.py module.
"""

from __future__ import absolute_import

import sys
import errno
import re
import logging
from time import time, sleep
import datetime
from random import randint
import requests
import pytest

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
from ..utils.pytest_extensions import simplified_test_function
pywbem = import_installed('pywbem')
from pywbem import WBEMListener  # noqa: E402
from pywbem._utils import _format  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

# Log level to be used. To enable logging, change this constant to the desired
# log level, and add code to set or unset the log level of the root logger
# globally or specifically in each test case.
LOGLEVEL = logging.NOTSET  # NOTSET disables logging

# Name of the log file
LOGFILE = 'test_indicationlistener.log'


def configure_root_logger(logfile):
    """
    Configure the root logger, except for log level
    """
    root_logger = logging.getLogger('')
    hdlr_exists = False
    for hdlr in root_logger.handlers:
        if isinstance(hdlr, logging.FileHandler):
            hdlr_exists = True
    if not hdlr_exists:
        hdlr = logging.FileHandler(logfile)
        hdlr.setFormatter(
            logging.Formatter(
                "%(asctime)s %(thread)s %(name)s %(levelname)s %(message)s"))
        root_logger.addHandler(hdlr)


LOGGER = logging.getLogger('test_indicationlistener')
if LOGLEVEL > logging.NOTSET:
    configure_root_logger(LOGFILE)


class ElapsedTimer(object):
    """
    Elapsed time timer.

    Calculates elapsed time between initiation/reset and access.
    """

    def __init__(self):
        """ Initiate the object with current time"""
        self.start_time = datetime.datetime.now()

    def reset(self):
        """ Reset the start time for the timer"""
        self.start_time = datetime.datetime.now()

    def elapsed_ms(self):
        """ Get the elapsed time in milliseconds. returns floating
            point representation of elapsed time in seconds.
        """
        dt = datetime.datetime.now() - self.start_time
        return ((dt.days * 24 * 3600) + dt.seconds) * 1000 + \
            dt.microseconds / 1000.0

    def elapsed_sec(self):
        """ get the elapsed time in seconds. Returns floating
            point representation of time in seconds
        """
        return self.elapsed_ms() / 1000


TESTCASES_WBEMLISTENER_INIT = [

    # Testcases for WBEMListener.__init__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to WBEMListener().
    #   * init_kwargs: Dict of keyword arguments to WBEMListener().
    #   * exp_attrs: Dict of expected attributes of resulting object.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Success cases
    (
        "Verify order of positional arguments",
        dict(
            init_args=[
                'woot.com',
                6997,
                6998,
                'certfile.pem',
                'keyfile.pem',
            ],
            init_kwargs={},
            exp_attrs=dict(
                host='woot.com',
                http_port=6997,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
        ),
        None, None, True
    ),
    (
        "Verify names of keyword arguments",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=6997,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=6997,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
        ),
        None, None, True
    ),
    (
        "Verify minimal arguments for success with http_port as integer",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=6997,
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=6997,
                https_port=None,
                certfile=None,
                keyfile=None,
            ),
        ),
        None, None, True
    ),
    (
        "Verify minimal arguments for success with http_port as string",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port='6997',
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=6997,
                https_port=None,
                certfile=None,
                keyfile=None,
            ),
        ),
        None, None, True
    ),
    (
        "Verify full arguments for success with http_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=6997,
                https_port=None,
                certfile=None,
                keyfile=None,
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=6997,
                https_port=None,
                certfile=None,
                keyfile=None,
            ),
        ),
        None, None, True
    ),
    (
        "Verify minimal arguments for success with https_port as integer",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
        ),
        None, None, True
    ),
    (
        "Verify minimal arguments for success with https_port as string",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                https_port='6998',
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
        ),
        None, None, True
    ),
    (
        "Verify full arguments for success with https_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
        ),
        None, None, True
    ),

    # Failure cases
    (
        "Verify failure when providing invalid type for http_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=12.1,
            ),
            exp_attrs=None,
        ),
        TypeError, None, True
    ),
    (
        "Verify failure when providing invalid type for https_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                https_port=12.1,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
            exp_attrs=None,
        ),
        TypeError, None, True
    ),
    (
        "Verify failure when providing no certfile with https_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                https_port=6998,
                certfile=None,
                keyfile='keyfile.pem',
            ),
            exp_attrs=None,
        ),
        ValueError, None, True
    ),
    (
        "Verify failure when providing no keyfile with https_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                https_port=6998,
                certfile='certfile.pem',
                keyfile=None,
            ),
            exp_attrs=None,
        ),
        ValueError, None, True
    ),
    (
        "Verify failure when providing no port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=None,
                https_port=None,
            ),
            exp_attrs=None,
        ),
        ValueError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_WBEMLISTENER_INIT)
@simplified_test_function
def test_WBEMListener_init(testcase, init_args, init_kwargs, exp_attrs):
    """
    Test function for WBEMListener.__init__()
    """

    # The code to be tested
    obj = WBEMListener(*init_args, **init_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify specified expected attributes
    for attr_name in exp_attrs:
        exp_attr = exp_attrs[attr_name]
        act_attr = getattr(obj, attr_name)
        assert act_attr == exp_attr
        assert isinstance(act_attr, type(exp_attr))

    # Verify attributes not set via init arguments

    assert isinstance(obj.logger, logging.Logger)
    assert re.match(r'pywbem\.listener\.', obj.logger.name)

    assert obj.http_started is False
    assert obj.https_started is False


TESTCASES_WBEMLISTENER_STR = [

    # Testcases for WBEMListener.__str__() / str()

    # Each list item is a testcase tuple with these items:
    # * obj: WBEMListener object to be tested.

    (
        WBEMListener(
            host='woot.com',
            http_port=6997,
            https_port=6998,
            certfile='certfile.pem',
            keyfile='keyfile.pem')
    ),
]


@pytest.mark.parametrize(
    "obj",
    TESTCASES_WBEMLISTENER_STR)
def test_WBEMListener_str(obj):
    """
    Test function for WBEMListener.__str__() / str()
    """

    # The code to be tested
    result = str(obj)

    assert re.match(r'^WBEMListener\(', result)

    exp_host_str = _format('_host={0!A}', obj.host)
    assert exp_host_str in result

    exp_http_port_str = _format('_http_port={0!A}', obj.http_port)
    assert exp_http_port_str in result

    exp_https_port_str = _format('_https_port={0!A}', obj.https_port)
    assert exp_https_port_str in result


TESTCASES_WBEMLISTENER_REPR = [

    # Testcases for WBEMListener.__repr__() / repr()

    # Each list item is a testcase tuple with these items:
    # * obj: WBEMListener object to be tested.

    (
        WBEMListener(
            host='woot.com',
            http_port=6997,
            https_port=6998,
            certfile='certfile.pem',
            keyfile='keyfile.pem')
    ),
]


@pytest.mark.parametrize(
    "obj",
    TESTCASES_WBEMLISTENER_REPR)
def test_WBEMListener_repr(obj):
    """
    Test function for WBEMListener.__repr__() / repr()
    """

    # The code to be tested
    result = repr(obj)

    assert re.match(r'^WBEMListener\(', result)

    exp_host_str = _format('_host={0!A}', obj.host)
    assert exp_host_str in result

    exp_http_port_str = _format('_http_port={0!A}', obj.http_port)
    assert exp_http_port_str in result

    exp_https_port_str = _format('_https_port={0!A}', obj.https_port)
    assert exp_https_port_str in result

    exp_certfile_str = _format('_certfile={0!A}', obj.certfile)
    assert exp_certfile_str in result

    exp_keyfile_str = _format('_keyfile={0!A}', obj.keyfile)
    assert exp_keyfile_str in result

    exp_logger_str = _format('_logger={0!A}', obj.logger)
    assert exp_logger_str in result


def test_WBEMListener_start_stop():
    """
    Test starting and stopping of the listener.
    """

    host = 'localhost'
    http_port = '50000'

    listener = WBEMListener(host, http_port)
    assert listener.http_started is False
    assert listener.https_started is False

    listener.start()
    assert listener.http_started is True
    assert listener.https_started is False

    listener.stop()
    assert listener.http_started is False
    assert listener.https_started is False


def test_WBEMListener_port_in_use():
    """
    Test starting the listener when port is in use by another listener.
    """

    host = 'localhost'

    # Don't use this port in other tests, to be on the safe side
    # as far as port reuse is concerned.
    http_port = '59999'

    exp_exc_type = OSError

    listener1 = WBEMListener(host, http_port)
    listener1.start()
    assert listener1.http_started is True

    listener2 = WBEMListener(host, http_port)

    try:

        # The code to be tested
        listener2.start()

    except Exception as exc:  # pylint: disable=broad-except
        # e.g. on Linux
        assert isinstance(exc, exp_exc_type)
        assert getattr(exc, 'errno', None) == errno.EADDRINUSE
        assert listener2.http_started is False
    else:
        # e.g. on Windows
        assert listener2.http_started is True

    # Verify that in any case, listener1 is still started
    assert listener1.http_started is True

    listener1.stop()  # cleanup
    listener2.stop()  # cleanup (for cases where it started)


def test_WBEMListener_context_mgr():
    """
    Test starting the listener and automatic closing in a context manager.
    """

    host = 'localhost'

    # Don't use this port in other tests, to be on the safe side
    # as far as port reuse is concerned.
    http_port = '59998'

    # The code to be tested (is the context manager)
    with WBEMListener(host, http_port) as listener1:

        # Verify that CM enter returns the listener
        assert isinstance(listener1, WBEMListener)

        listener1.start()
        assert listener1.http_started is True

    # Verify that CM exit stops the listener
    assert listener1.http_started is False

    # Verify that the TCP/IP port can be used again
    listener2 = WBEMListener(host, http_port)
    listener2.start()
    assert listener2.http_started is True
    listener2.stop()


def create_indication_data(msg_id, sequence_number, delta_time, protocol_ver):
    """
    Create a test indication from the template and input attributes.
    """

    data_template = """<?xml version="1.0" encoding="utf-8" ?>
    <CIM CIMVERSION="2.0" DTDVERSION="2.4">
      <MESSAGE ID="{msg_id}" PROTOCOLVERSION="{protocol_ver}">
        <SIMPLEEXPREQ>
          <EXPMETHODCALL NAME="ExportIndication">
            <EXPPARAMVALUE NAME="NewIndication">
              <INSTANCE CLASSNAME="CIM_AlertIndication">
                <PROPERTY NAME="Severity" TYPE="string">
                  <VALUE>high</VALUE>
                </PROPERTY>
                <PROPERTY NAME="SequenceNumber" TYPE="string">
                  <VALUE>{sequence_number}</VALUE>
                </PROPERTY>
                <PROPERTY NAME="DELTA_TIME" TYPE="string">
                  <VALUE>{delta_time}</VALUE>
                </PROPERTY>
              </INSTANCE>
            </EXPPARAMVALUE>
          </EXPMETHODCALL>
        </SIMPLEEXPREQ>
      </MESSAGE>
    </CIM>"""

    return data_template.format(
        sequence_number=sequence_number, delta_time=delta_time,
        protocol_ver=protocol_ver, msg_id=msg_id)


# Verbosity in test_WBEMListener_send_indications()
VERBOSE_SUMMARY = False  # Show summary for each run

# Global variables used to communicate between the test case function and
# the process_indication() function running in context of the listener
# thread. These must be global, because in Python 2, closure variables
# cannot be modified.
RCV_COUNT = 0
RCV_ERRORS = False


def process_indication(indication, host):
    # pylint: disable=unused-argument
    """
    This function gets called by the listener when an indication is
    received.

    It tests the received indication sequence number against the RCV_COUNT
    which should catch any duplicated or missing indication, since the
    counters would no longer match in such cases.

    This function is invoked in context of the listener thread. Therefore,
    it is defined as a local function in order to be able to use the
    variables from its outer function, specifically the RCV_COUNT and
    RCV_ERRORS variables.

    For the same reason, this function does not report assertion failures
    by means of raising exceptions, but by printing a message and setting
    RCV_ERRORS to an error message, so the test function can check it and
    raise assertion errors.
    """

    # Note: Global variables that are modified must be declared global
    global RCV_COUNT  # pylint: disable=global-statement
    global RCV_ERRORS  # pylint: disable=global-statement

    LOGGER.debug(
        "Callback function called for indication #%s: %r",
        RCV_COUNT, indication.classname)

    try:

        send_count = int(indication.properties['SequenceNumber'].value)
        if send_count != RCV_COUNT:
            print("Error in process_indication(): Assertion error: "
                  "Unexpected SequenceNumber in received indication: "
                  "got {}, expected {}".format(send_count, RCV_COUNT))
            sys.stdout.flush()
            RCV_ERRORS = True

        RCV_COUNT += 1

    except Exception as exc:  # pylint: disable=broad-except
        print("Error in process_indication(): {}: {}".
              format(exc.__class__.__name__, exc))
        sys.stdout.flush()
        RCV_ERRORS = True


@pytest.mark.parametrize(
    "send_count",
    [1, 10, 100]  # 1000 in some environments takes 30 min
)
def test_WBEMListener_send_indications(send_count):
    """
    Test WBEMListener with an indication generator.

    This test sends the number of indications defined by the send_count
    parameter using HTTP. It confirms that they are all received by the
    listener.

    This test validates the main paths of the listener and that the listener can
    receive large numbers of indications without duplicates or dropping
    indications.

    It does not validate all of the possible xml options on indications.

    Creates the listener, starts the listener, creates the indication XML and
    adds sequence number and time to the indication instance and sends that
    instance using requests.

    The indication instance is modified for each indication count so that each
    carries its own sequence number.
    """

    # Note: Global variables that are modified must be declared global
    global RCV_COUNT  # pylint: disable=global-statement
    global RCV_ERRORS  # pylint: disable=global-statement

    # Enable logging for this test function
    if LOGLEVEL > logging.NOTSET:
        logging.getLogger('').setLevel(LOGLEVEL)

    host = 'localhost'
    http_port = 50000

    listener = WBEMListener(host, http_port)
    listener.add_callback(process_indication)
    listener.start()

    try:

        start_time = time()
        url = 'http://{}:{}'.format(host, http_port)
        cim_protocol_version = '1.4'
        headers = {
            'Content-Type': 'application/xml; charset=utf-8',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': cim_protocol_version,
        }
        # We include Accept-Encoding because of requests issue.
        # He supplies it if we don't.  TODO try None

        delta_time = time() - start_time
        random_base = randint(1, 10000)
        timer = ElapsedTimer()

        RCV_COUNT = 0
        RCV_ERRORS = False

        for i in range(send_count):

            msg_id = random_base + i
            payload = create_indication_data(msg_id, i, delta_time,
                                             cim_protocol_version)

            LOGGER.debug("Testcase sending indication #%s", i)

            try:
                response = requests.post(url, headers=headers, data=payload,
                                         timeout=4)
            except requests.exceptions.RequestException as exc:
                msg = ("Testcase sending indication #{} raised {}: {}".
                       format(i, exc.__class__.__name__, exc))
                LOGGER.error(msg)
                new_exc = AssertionError(msg)
                new_exc.__cause__ = None  # Disable to see original traceback
                raise new_exc

            LOGGER.debug("Testcase received response from sending "
                         "indication #%s", i)

            if response.status_code != 200:
                msg = ("Testcase sending indication #{} failed with HTTP "
                       "status {}".format(i, response.status_code))
                LOGGER.error(msg)
                raise AssertionError(msg)

        endtime = timer.elapsed_sec()

        # Make sure the listener thread has processed all indications
        sleep(1)

        if VERBOSE_SUMMARY:
            print("\nSent {} indications in {} sec or {:.2f} ind/sec".
                  format(send_count, endtime, (send_count / endtime)))
            sys.stdout.flush()

        assert not RCV_ERRORS, \
            "Errors occurred in process_indication(), as printed to stdout"

        assert send_count == RCV_COUNT, \
            "Mismatch between total send count {} and receive count {}". \
            format(send_count, RCV_COUNT)

    finally:
        listener.stop()

        # Disable logging for this test function
        if LOGLEVEL > logging.NOTSET:
            logging.getLogger('').setLevel(logging.NOTSET)


@pytest.mark.parametrize(
    "method, exp_status", [
        ('OPTIONS', 405),
        ('HEAD', 405),
        ('GET', 405),
        ('PUT', 405),
        ('PATCH', 405),
        ('DELETE', 405),
        ('TRACE', 405),
        ('CONNECT', 405),
        ('M_POST', 405),
    ])
def test_WBEMListener_incorrect_method(method, exp_status):
    """
    Verify that WBEMListener send fails when an incorrect HTTP method is used.
    """

    host = 'localhost'
    http_port = 50000
    url = 'http://{}:{}'.format(host, http_port)
    headers = {
        'Content-Type': 'application/xml; charset=utf-8',
        'CIMExport': 'MethodRequest',
        'CIMExportMethod': 'ExportIndication',
        'Accept-Encoding': 'Identity',
        'CIMProtocolVersion': '1.4',
    }

    listener = WBEMListener(host, http_port)
    listener.add_callback(process_indication)
    listener.start()

    try:

        # The code to be tested is running in listener thread
        response = requests.request(method, url, headers=headers, timeout=4)

        assert response.status_code == exp_status

    finally:
        listener.stop()


WBEMLISTENER_INCORRECT_HEADERS_TESTCASES = [
    (
        "Invalid Accept-Charset header",
        {
            'Content-Type': 'application/xml',
            'Accept-Charset': 'ASCII',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Invalid Accept-Charset header .*',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Invalid Accept-Encoding header",
        {
            'Content-Type': 'application/xml',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'foo',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Invalid Accept-Encoding header .*',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Accept-Range header not permitted",
        {
            'Content-Type': 'application/xml',
            'Accept-Range': 'foo',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Accept-Range header is not permitted .*',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Missing Content-Type header",
        {
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Content-Type header is required',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Invalid Content-Type header",
        {
            'Content-Type': 'foo_application/xml',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Invalid Content-Type header .*',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Invalid Content-Encoding header",
        {
            'Content-Type': 'application/xml',
            'Content-Encoding': 'foo',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Invalid Content-Encoding header .*',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
]


@pytest.mark.parametrize(
    "desc, headers, exp_status, exp_headers",
    WBEMLISTENER_INCORRECT_HEADERS_TESTCASES)
def test_WBEMListener_incorrect_headers(desc, headers, exp_status, exp_headers):
    # pylint: disable=unused-argument
    """
    Verify that WBEMListener send fails when incorrect HTTP headers are used
    (along with the correct POST method).
    """

    host = 'localhost'
    http_port = 50000
    url = 'http://{}:{}'.format(host, http_port)
    # headers = copy(headers)

    listener = WBEMListener(host, http_port)
    listener.add_callback(process_indication)
    listener.start()

    try:

        # The code to be tested is running in listener thread
        response = requests.post(url, headers=headers, timeout=4)

        assert response.status_code == exp_status
        for header_name in exp_headers:
            assert header_name in response.headers
            exp_header_pattern = exp_headers[header_name]
            assert re.match(exp_header_pattern, response.headers[header_name])

    finally:
        listener.stop()


WBEMLISTENER_INCORRECT_PAYLOAD1_TESTCASES = [
    (
        "Ill-formed XML in payload",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">""",
        400,
        {
            'CIMErrorDetails': r'XML parsing error .*',
            'CIMError': r'request-not-well-formed',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Unsupported DTD version (DTDVERSION in CIM element)",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="1.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'DTD version 1.4 not supported.*',
            'CIMError': r'unsupported-dtd-version',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Unsupported protocol version (PROTOCOLVERSION in MESSAGE element)",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="2.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'Protocol version 2.4 not supported.*',
            'CIMError': r'unsupported-protocol-version',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "CIM-XML: Missing MESSAGE element",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'Element .CIM. missing required '
                               r'child element .*.MESSAGE..*',
            'CIMError': r'request-not-well-formed',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "CIM-XML: Missing SIMPLEEXPREQ element",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
          </MESSAGE>
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'Element .MESSAGE. missing required '
                               r'child element .*.SIMPLEEXPREQ..*',
            'CIMError': r'request-not-well-formed',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "CIM-XML: Missing EXPMETHODCALL element",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'Element .SIMPLEEXPREQ. missing required '
                               r'child element .*.EXPMETHODCALL..*',
            'CIMError': r'request-not-well-formed',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "CIM-XML: Invalid parameter type in ExportIndication export request",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
                <EXPPARAMVALUE NAME="NewIndication">
                  <INSTANCENAME CLASSNAME="CIM_AlertIndication">
                  </INSTANCENAME>
                </EXPPARAMVALUE>
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'Element .EXPPARAMVALUE. has invalid child '
                               r'element.*INSTANCENAME.*',
            'CIMError': r'request-not-well-formed',
            'CIMExport': r'MethodResponse',
        }
    ),
]


@pytest.mark.parametrize(
    "desc, payload, exp_status, exp_headers",
    WBEMLISTENER_INCORRECT_PAYLOAD1_TESTCASES)
def test_WBEMListener_incorrect_payload1(
        desc, payload, exp_status, exp_headers):
    # pylint: disable=unused-argument
    """
    Verify that WBEMListener send fails with HTTP error when incorrect HTTP
    payload is used that triggers HTTP errors.
    """

    host = 'localhost'
    http_port = 50000
    url = 'http://{}:{}'.format(host, http_port)
    headers = {
        'Content-Type': 'application/xml; charset=utf-8',
        'CIMExport': 'MethodRequest',
        'CIMExportMethod': 'ExportIndication',
        'Accept-Encoding': 'Identity',
        'CIMProtocolVersion': '1.4',
    }

    listener = WBEMListener(host, http_port)
    listener.add_callback(process_indication)
    listener.start()

    try:

        # The code to be tested is running in listener thread
        response = requests.post(url, headers=headers, data=payload, timeout=4)

        assert response.status_code == exp_status
        for header_name in exp_headers:
            assert header_name in response.headers
            exp_header_pattern = exp_headers[header_name]
            assert re.match(exp_header_pattern, response.headers[header_name])

    finally:
        listener.stop()


WBEMLISTENER_INCORRECT_PAYLOAD2_TESTCASES = [
    (
        "Unsupported export method in export request",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="fooExportIndication">
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        200,
        {
            'CIMExport': r'MethodResponse',
        },
        b'<?xml version="1.0" encoding="utf-8" ?>\n'
        b'<CIM CIMVERSION="2.0" DTDVERSION="2.4">'
        b'<MESSAGE ID="42" PROTOCOLVERSION="1.4">'
        b'<SIMPLEEXPRSP>'
        b'<EXPMETHODRESPONSE NAME="fooExportIndication">'
        b'<ERROR CODE="7" DESCRIPTION="Unknown export method: '
        b'\'fooExportIndication\'"/>'
        b'</EXPMETHODRESPONSE>'
        b'</SIMPLEEXPRSP>'
        b'</MESSAGE>'
        b'</CIM>'
    ),
    (
        "No parameters in ExportIndication export request",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        200,
        {
            'CIMExport': r'MethodResponse',
        },
        b'<?xml version="1.0" encoding="utf-8" ?>\n'
        b'<CIM CIMVERSION="2.0" DTDVERSION="2.4">'
        b'<MESSAGE ID="42" PROTOCOLVERSION="1.4">'
        b'<SIMPLEEXPRSP>'
        b'<EXPMETHODRESPONSE NAME="ExportIndication">'
        b'<ERROR CODE="7" DESCRIPTION="Expecting one parameter '
        b'NewIndication.*"/>'
        b'</EXPMETHODRESPONSE>'
        b'</SIMPLEEXPRSP>'
        b'</MESSAGE>'
        b'</CIM>'
    ),
    (
        "Two parameters in ExportIndication export request",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
                <EXPPARAMVALUE NAME="NewIndication">
                  <INSTANCE CLASSNAME="CIM_AlertIndication">
                  </INSTANCE>
                </EXPPARAMVALUE>
                <EXPPARAMVALUE NAME="foo">
                  <INSTANCE CLASSNAME="CIM_AlertIndication">
                  </INSTANCE>
                </EXPPARAMVALUE>
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        200,
        {
            'CIMExport': r'MethodResponse',
        },
        b'<?xml version="1.0" encoding="utf-8" ?>\n'
        b'<CIM CIMVERSION="2.0" DTDVERSION="2.4">'
        b'<MESSAGE ID="42" PROTOCOLVERSION="1.4">'
        b'<SIMPLEEXPRSP>'
        b'<EXPMETHODRESPONSE NAME="ExportIndication">'
        b'<ERROR CODE="7" DESCRIPTION="Expecting one parameter '
        b'NewIndication.*"/>'
        b'</EXPMETHODRESPONSE>'
        b'</SIMPLEEXPRSP>'
        b'</MESSAGE>'
        b'</CIM>'
    ),
    (
        "Invalid parameter name in ExportIndication export request",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
                <EXPPARAMVALUE NAME="fooNewIndication">
                  <INSTANCE CLASSNAME="CIM_AlertIndication">
                  </INSTANCE>
                </EXPPARAMVALUE>
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        200,
        {
            'CIMExport': r'MethodResponse',
        },
        b'<?xml version="1.0" encoding="utf-8" ?>\n'
        b'<CIM CIMVERSION="2.0" DTDVERSION="2.4">'
        b'<MESSAGE ID="42" PROTOCOLVERSION="1.4">'
        b'<SIMPLEEXPRSP>'
        b'<EXPMETHODRESPONSE NAME="ExportIndication">'
        b'<ERROR CODE="7" DESCRIPTION="Expecting one parameter '
        b'NewIndication.*"/>'
        b'</EXPMETHODRESPONSE>'
        b'</SIMPLEEXPRSP>'
        b'</MESSAGE>'
        b'</CIM>'
    ),
]


@pytest.mark.parametrize(
    "desc, payload, exp_status, exp_headers, exp_payload",
    WBEMLISTENER_INCORRECT_PAYLOAD2_TESTCASES)
def test_WBEMListener_incorrect_payload2(
        desc, payload, exp_status, exp_headers, exp_payload):
    # pylint: disable=unused-argument
    """
    Verify that WBEMListener send fails with export response indicating error
    when incorrect HTTP payload is used that triggers that.
    """

    host = 'localhost'
    http_port = 50000
    url = 'http://{}:{}'.format(host, http_port)
    headers = {
        'Content-Type': 'application/xml; charset=utf-8',
        'CIMExport': 'MethodRequest',
        'CIMExportMethod': 'ExportIndication',
        'Accept-Encoding': 'Identity',
        'CIMProtocolVersion': '1.4',
    }

    listener = WBEMListener(host, http_port)
    listener.add_callback(process_indication)
    listener.start()

    try:

        # The code to be tested is running in listener thread
        response = requests.post(url, headers=headers, data=payload, timeout=4)

        assert response.status_code == exp_status
        for header_name in exp_headers:
            assert header_name in response.headers
            exp_header_pattern = exp_headers[header_name]
            assert re.match(exp_header_pattern, response.headers[header_name])
        act_payload = response.content
        re.match(exp_payload, act_payload, re.MULTILINE)

    finally:
        listener.stop()



#!/usr/bin/env python
"""
Test cases for _tupletree module and Unicode/XML check functions.
"""

from __future__ import absolute_import, print_function

import sys as _sys
import os as _os
import time
import argparse as _argparse
import datetime
import pstats
import cProfile
import locale
from tabulate import tabulate
from pyinstrument import Profiler
import six

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from tests.utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import _tupletree, __version__  # noqa: E402
from pywbem._cliutils import SmartFormatter as _SmartFormatter  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

# In pywbem 0.13, parse_cim() changed from a function to a method:
try:
    from pywbem._tupleparse import parse_cim
except ImportError:
    from pywbem._tupleparse import TupleParser

    def parse_cim(tt):
        """Compatible parse_cim() function"""
        tp = TupleParser()
        return tp.parse_cim(tt)

if six.PY2:
    import codecs  # pylint: disable=wrong-import-order

# default input arguments.  These are small so a user who
# just calls the code does not fall into long run.
DEFAULT_RESPONSE_SIZE = [100, 1000]
DEFAULT_RESPONSE_COUNT = [100, 1000]
DEFAULT_TOP_N_ROWS = 20
DEFAULT_RUNID = 'default'


PROFILE_OUT_PREFIX = 'perf'
PROFILE_DUMP_SUFFIX = 'profile'
PROFILE_LOG_SUFFIX = 'out'

STDOUT_ENCODING = getattr(_sys.stdout, 'encoding', None)
if not STDOUT_ENCODING:
    STDOUT_ENCODING = locale.getpreferredencoding()
if not STDOUT_ENCODING:
    STDOUT_ENCODING = 'utf-8'


def _uprint(dest, text):  # pylint: disable=too-many-branches
    """
    Write text to dest, adding a newline character.

    Text may be a unicode string, or a byte string in UTF-8 encoding.
    It must not be None.

    If dest is None, the text is encoded to a codepage suitable for the current
    stdout and is written to stdout.

    Otherwise, dest must be a file path, and the text is encoded to a UTF-8
    Byte sequence and is appended to the file (opening and closing the file).
    """
    if isinstance(text, six.text_type):
        text = text + u'\n'
    elif isinstance(text, six.binary_type):
        text = text + b'\n'
    else:
        raise TypeError(
            "text must be a unicode or byte string, but is {0}".
            format(type(text)))
    if dest is None:
        if six.PY2:
            # On py2, stdout.write() requires byte strings
            if isinstance(text, six.text_type):
                text = text.encode(STDOUT_ENCODING, 'replace')
        else:
            # On py3, stdout.write() requires unicode strings
            if isinstance(text, six.binary_type):
                text = text.decode('utf-8')
        _sys.stdout.write(text)

    elif isinstance(dest, (six.text_type, six.binary_type)):
        if isinstance(text, six.text_type):
            open_kwargs = dict(mode='a', encoding='utf-8')
        else:
            open_kwargs = dict(mode='ab')
        if six.PY2:
            # Open with codecs to be able to set text mode
            with codecs.open(dest, **open_kwargs) as fn:
                fn.write(text)
        else:
            with open(dest, **open_kwargs) as fn:
                fn.write(text)
    else:
        raise TypeError(
            "dest must be None or a string, but is {0}".
            format(type(text)))


class _PywbemCustomFormatter(_SmartFormatter,
                             _argparse.RawDescriptionHelpFormatter):
    """
    Define a custom Formatter to allow formatting help and epilog.

    argparse formatter specifically allows multiple inheritance for the
    formatter customization and actually recommends this in a discussion
    in one of the issues:

        https://bugs.python.org/issue13023

    Also recommended in a StackOverflow discussion:

    https://stackoverflow.com/questions/18462610/argumentparser-epilog-and-description-formatting-in-conjunction-with-argumentdef
    """
    pass  # pylint: disable=unnecessary-pass


TEMPLATE = u"""\n<VALUE.NAMEDINSTANCE>
<INSTANCENAME CLASSNAME="TST_ResponseStressTestCxx">
  <KEYBINDING NAME="Id"><KEYVALUE VALUETYPE="string">{0}</KEYVALUE>
    </KEYBINDING></INSTANCENAME><INSTANCE CLASSNAME="PERFTEST_Class" >
  <PROPERTY NAME="Id" TYPE="string"><VALUE>{0}</VALUE></PROPERTY>
  <PROPERTY NAME="SequenceNumber"  TYPE="uint64"><VALUE>{0}</VALUE></PROPERTY>
  <PROPERTY NAME="S1"  TYPE="string"><VALUE>{2}</VALUE></PROPERTY>
  <PROPERTY.ARRAY NAME="A1" TYPE="string">
    <VALUE.ARRAY>
        {3}
    </VALUE.ARRAY>
  </PROPERTY.ARRAY>
</INSTANCE></VALUE.NAMEDINSTANCE>
"""
AVG_OBJ_SIZE = 0


def build_xml_obj(number, size):
    """
        Builds a single XML object
    """
    if size > 700:
        pattern = "abcdefghighjklmnopqrstuvwxyz01234567890"
    else:
        pattern = 'ab'

    # compute sizes for the s1_value and a1_value that will create an xml
    # of about the size defined by size

    # NOTE: This creates a minimum XML size of about 600 bytes just with the
    # overhead
    fill_size = size - len(TEMPLATE)
    if fill_size <= 0:
        fill_repeat = 1
    else:
        fill_repeat = int((fill_size / len(pattern)) / 2)

    s1_value = pattern * fill_repeat
    a1_value = "<VALUE>{0}</VALUE>\n".format(pattern) * fill_repeat

    return TEMPLATE.format(number, pattern, s1_value, a1_value)


AVG_RESPONSE_SIZE = 0


def create_xml_objs(count, size):
    """
    Create the set of CIMInstances XML that will go into the test XML
    """
    obj_size = []
    objstr = ""
    for number in range(0, count):
        obj = build_xml_obj(number, size)
        obj_size.append(len(obj))
        objstr += obj
    global AVG_RESPONSE_SIZE  # pylint: disable=global-statement
    AVG_RESPONSE_SIZE = sum(obj_size) / len(obj_size)

    return objstr


def create_xml(count=1000, size=100):
    """
    Create the xml response message
    """
    xml = u'<?xml version="1.0" encoding="utf-8" ?>' \
          '<CIM CIMVERSION="2.0" DTDVERSION="2.0">' \
          '<MESSAGE ID="1001" PROTOCOLVERSION="1.0"><SIMPLERSP>' \
          '<IMETHODRESPONSE NAME="EnumerateInstances"><IRETURNVALUE>'

    xml += create_xml_objs(count=count, size=size)

    xml += '\n</IRETURNVALUE></IMETHODRESPONSE></SIMPLERSP></MESSAGE></CIM>'

    return xml, AVG_RESPONSE_SIZE


def execute_test_code(xml_string, profiler):
    """
    The test code to be executed. If a profiler is defined it is enabled
    just before the test code is executed and disabled just after the
    code is executed.
    """
    if profiler:
        if isinstance(profiler, cProfile.Profile):
            profiler.enable()
        elif isinstance(profiler, Profiler):
            profiler.start()

    # The code to be tested
    tt = _tupletree.xml_to_tupletree_sax(xml_string, "TestData")

    parse_cim(tt)

    if profiler:
        if isinstance(profiler, cProfile.Profile):
            profiler.disable()
        elif isinstance(profiler, Profiler):
            profiler.stop()


def execute_with_time(xml_string, profiler):
    # desc reserved for future tests.
    """
    Start time measurement and execute the test code.

    Returns the execution time
    """
    start_time = time.time()

    execute_test_code(xml_string, profiler)

    return time.time() - start_time


class ExecuteTests(object):
    # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """
    Params class contains args parameters but lets us create new parameters
    or modify the existing paramete values.  argparse does not allow modifing
    created argument values.
    """
    def __init__(self, args):
        self.response_size = args.response_size
        self.response_count = args.response_count
        self.profiler = args.profiler
        self.verbose = args.verbose
        self.log = args.log
        self.runid = args.runid or DEFAULT_RUNID
        self.runid = self.runid + ": " + __version__
        # insure no blanks in the text
        self.runid.replace(" ", "-")
        self.tbl_output_format = 'simple'

        self.file_datetime = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        # set logfile name if log option set. If the individual
        # option set, this will be modified
        if self.log:
            self.logfile = "{0}_{1}_{2}_{3}.{4}".format(
                PROFILE_OUT_PREFIX,
                self.runid,
                self.profiler,
                self.file_datetime,
                PROFILE_LOG_SUFFIX)
        else:
            self.logfile = None

        # define a dumpfile for the cprofile save.
        self.dumpfilename = "{0}_{1}_{2}_{3}.{4}".format(
            PROFILE_OUT_PREFIX,
            self.runid,
            self.profiler,
            self.file_datetime,
            PROFILE_DUMP_SUFFIX)

        # number of lines to display for cprofile output
        self.top_n_rows = args.top_n_rows

        # The following describes the columns in the cprofile output.
        # ncalls - how many times the function/method has been called (in case
        # the same function/method is being called recursively then ncalls has
        # two values eg. 120/20, where the first is the true number of calls,
        # while the second is the number of direct calls)

        # tottime - the total time in seconds excluding the time of other
        # functions/methods

        # percall - average time to execute function (per call)

        # cumtime - the total time in seconds includes the times of other
        # functions it calls

        self.cprofilesort = ["tottime", "ncalls"]

    def __repr__(self):
        return "ExecuteTests(runid={0}, response_size={1} response_count={2} " \
               "profiler={3} verbose={4}, log={5}, file_datetime={6}, " \
               "logfile={7}, dumpfilename={8}, top_n_rows={9}, " \
               "cprofilesort={10})".format(
                   self.runid,
                   self.response_size,
                   self.response_count,
                   self.profiler,
                   self.verbose,
                   self.log,
                   self.file_datetime,
                   self.logfile,
                   self.dumpfilename,
                   self.top_n_rows,
                   self.cprofilesort)

    def execute_raw_tests(self, profiler=None):
        """
        Execute the parse test for all of the input parameters defined in in
        self.response_count and self.response_size. This allows multiple tests
        to be executed
        """
        table_rows = []
        for response_size in self.response_size:
            for response_count in self.response_count:
                xml, avg_resp_size = create_xml(response_count, response_size)
                execution_time = execute_with_time(xml, profiler=profiler)
                row = (response_size,
                       int(avg_resp_size),
                       response_count,
                       len(xml),
                       "{0:5.2f}".format(execution_time),
                       int(response_count / execution_time))
                table_rows.append(row)
        return table_rows

    def execute_cprofile_tests(self):
        """
        Execute the parse test for all of the input parameters. The cProfiler
        allows the profiling to be enabled and disabled so the profiler object
        is passed on to the execution component
        """

        profiler = cProfile.Profile()

        table_rows = self.execute_raw_tests(profiler=profiler)

        # Define stats output.
        ps = pstats.Stats(profiler, stream=_sys.stdout)

        try:
            ps.dump_stats(self.dumpfilename)
        except Exception as ex:  # pylint: disable=broad-except
            print('cProfiler dump stats exception %s %s. Ignored' %
                  (ex.__class__.__name__, ex))

        # modify stats for display and output to stdout
        ps.strip_dirs()
        ps.sort_stats(*self.cprofilesort)
        ps.print_stats(self.top_n_rows)

        # if dest defined, output to file defined by dest
        if self.logfile:
            with open(self.logfile, 'w') as stream:
                ps = pstats.Stats(profiler, stream=stream)
                ps.strip_dirs()
                ps.sort_stats(*self.cprofilesort)
                ps.print_stats(self.top_n_rows)

        return table_rows

    def execute_pyinstrument_tests(self):
        """Execute the parse test for all of the input parameters in args.
        Since this profiler has no enable or disable concept the profiler
        must be enabled for the complete test and the results output
        At the end of the test, the profiler results are printed
        """
        table_rows = []
        profiler = Profiler()

        table_rows = self.execute_raw_tests(profiler)

        _uprint(None, profiler.output_text(unicode=True, color=True))
        if self.logfile:
            _uprint(self.logfile, profiler.output_text(unicode=True,
                                                       color=True))
        return table_rows

    def execute_tests(self):
        """Execute the test associated with profiler input argument."""

        if self.profiler == 'none':
            table_rows = self.execute_raw_tests()
        elif self.profiler == 'pyinst':
            table_rows = self.execute_pyinstrument_tests()
        elif self.profiler == 'cprofile':
            table_rows = self.execute_cprofile_tests()
        else:
            raise RuntimeError('profiler arg %s invalid. '
                               'Should never occur' % self.profiler)

        # build and output results report
        header = ["Exp inst\nsize",
                  "Act inst\nsize",
                  "Response\nCount",
                  "XML\nsize",
                  "Parse time\nsec.",
                  "Instances\nper sec"]
        title = 'Results: profile={0}, response_counts={1},\n   ' \
                'response-sizes={2}, runid={3} {4}'.format(self.profiler,
                                                           self.response_count,
                                                           self.response_size,
                                                           self.runid,
                                                           self.file_datetime)
        table = tabulate(table_rows, header, tablefmt=self.tbl_output_format)

        # print statistics to terminal
        print("")
        _uprint(None, title)
        _uprint(None, table)

        # print statistics to log file
        if self.logfile:
            _uprint(self.logfile, title)
            _uprint(self.logfile, table)


def execute_individual_tests(args):
    """
    Execute the test with separate trace for each test.  This  creates
    a Params object for each set of response_size and response_count so
    that  tests can be executed completely individually.
    """
    # create a params object
    tests = ExecuteTests(args)

    # modify the test specific variables of the params and execute the test
    # with each
    for response_size in args.response_size:
        for response_count in args.response_count:
            # define a dump file name for each execution
            tests.response_count = [response_count]
            tests.response_size = [response_size]
            tests.dumpfilename = "{0}_{1}_{2}_{3}_{4}_{5}.{6}".format(
                PROFILE_OUT_PREFIX,
                tests.runid,
                args.profiler,
                tests.file_datetime,
                response_count,
                response_size,
                PROFILE_DUMP_SUFFIX)

            # If a log destination is defined modify the dump file name
            # for each individual  test suffixes the name with the
            # response_size and response_count
            if args.log:
                tests.logfile = "{0}_{1}_{2}_{3}_{4}_{5}.{6}".format(
                    PROFILE_OUT_PREFIX,
                    tests.runid,
                    args.profiler,
                    tests.file_datetime,
                    response_count,
                    response_size,
                    PROFILE_LOG_SUFFIX)

            tests.execute_tests()


def parse_args():
    """
    Parse the input arguments and return the args dictionary
    """
    prog = _os.path.basename(_sys.argv[0])
    usage = '%(prog)s [options]'
    # pylint: disable=line-too-long
    desc = """
This script provide performance information on the TupleParse class.  This  is
a development test to be used to improve this code and reduce the XML response
execution time.

It creates XML for an EnumerateInstances response with the number of
instances (response-count) and the approximate XML size of each instance
(response-size) defined by the input arguments.  It then executes the parsing
sequence (_tupletree, _tupleparse) against this xml,records the execution time.

The input arguments for response count and response size may each specify
multiple values (ex --response-count 100 100 1000). In this case tests are
executed for all variations of each of the arguments.

If a profiler is specified (--profiler input parameter) it executes the
profile for all of the tests and displays the profile results before outputting
the execution time for each test.

NOTE: Running with a profiler heavily affects the runtime.
"""
    epilog = """
Examples:
  %s

     Execute a minimal test with default input arguments and display time to
     execute. It does not use either of the profilers.

  %s  -p cprofile --response-count 10000 20000 \\
    --response-size 100 1000

     Execute the test for all combinations of response counts of 10,000 and
     20,000 and for response sizes of 500 and 1000 bytes using cprofile
     profiler to generate and output a profile of the operation.
""" % (prog, prog)  # noqa: E501
# pylint: enable=line-too-long

    argparser = _argparse.ArgumentParser(
        prog=prog, usage=usage, description=desc, epilog=epilog,
        add_help=False, formatter_class=_PywbemCustomFormatter)

    tests_arggroup = argparser.add_argument_group(
        'Test related options',
        'Specify parameters of the test')

    tests_arggroup.add_argument(
        '-p', '--profiler',
        dest='profiler', choices=['none', 'pyinst', 'cprofile'],
        action='store', default='none',
        help='R|Defines the profile package used for the test.\n'
             '   * `pyinst` uses pyinstrument uses a statistical \n'
             '     capture, and displays a tree of the python stack\n'
             '     execution times.\n'
             '   * `cprofile` uses cProfile and generates a table of\n'
             '     counts.\n'
             '   * `none` runs without profiler.\n'
             ' Default: %s' % "none")

    tests_arggroup.add_argument(
        '-c', '--response-count', dest='response_count', nargs='+',
        metavar='ints', type=int,
        action='store', default=DEFAULT_RESPONSE_COUNT,
        help='R|The number of instances that will be returned for each\n'
             'test in the form for each test. May be multiple \n'
             'integers. The test will be executed for each\n'
             'response-size and each value defined. The format is:\n'
             '   -c 1000 10000 100000\n'
             'Default: %s' % DEFAULT_RESPONSE_COUNT)

    tests_arggroup.add_argument(
        '-s', '--response-size', dest='response_size', nargs='+',
        metavar='ints', type=int,
        action='store', default=DEFAULT_RESPONSE_SIZE,
        help='R|The response sizes that will be tested. This defines\n'
             'the size of the XML for each object in the response\n'
             'in bytes. May be multiple integers. the test will be\n'
             'executed for the combination of each response-count\n'
             'and each value provided. The format is:\n'
             '   -s 100 200 300\n'
             'Default: %s' % DEFAULT_RESPONSE_SIZE)

    tests_arggroup.add_argument(
        '-n', '--top-n-rows', dest='top_n_rows',
        metavar='int', type=int,
        action='store', default=DEFAULT_TOP_N_ROWS,
        help='R|The number of rows of profile data for the cprofile\n'
             'display. This is the top n tottime results.\n'
             'Default: %s' % DEFAULT_TOP_N_ROWS)

    tests_arggroup.add_argument(
        '-r', '--runid', dest='runid',
        metavar='text', type=str,
        action='store', default=DEFAULT_RUNID,
        help='R|A string that is concatenated into each filename\n'
             'created by the script. It is suffixed with pywbem version.\n'
             'Default: %s' % DEFAULT_RUNID)

    tests_arggroup.add_argument(
        '-i', '--individual', dest='individual',
        action='store_true', default=False,
        help='R|Run each of the response_count/response_size tests as a\n'
             'completely individual test with separate profile and separate\n'
             'output displays and log files.')

    general_arggroup = argparser.add_argument_group(
        'General options')

    general_arggroup.add_argument(
        '-l', '--log', dest='log',
        action='store_true', default=False,
        help='R|Log the trace for each either the complete execution or each\n'
             'test if --individual set to a log file.')

    general_arggroup.add_argument(
        '-v', '--verbose', dest='verbose',
        action='store_true', default=False,
        help='R|Print more messages while processing. Displays detailed\n'
             'counts for each pull operation.')
    general_arggroup.add_argument(
        '-h', '--help', action='help',
        help='Show this help message and exit')

    args = argparser.parse_args()

    if args.top_n_rows <= 0:
        argparser.error("top-n-rows must be postive integer. "
                        "{0} not allowed".format(args.top_n_rows))

    return args


def main():
    """
        Parse arguments, execute tests defined by the arguments, and
        display results.
    """
    args = parse_args()

    print('Starting profiler={0}, response-count={1} response-size={2} '
          'runid={3} log={4}'.format(args.profiler, args.response_count,
                                     args.response_size, args.runid,
                                     args.log))

    if args.individual:
        execute_individual_tests(args)
    else:
        tests = ExecuteTests(args)
        tests.execute_tests()


if __name__ == '__main__':
    main()



"""
Test memory leaks for xml.etree.ElementTree classes.
"""

from __future__ import absolute_import, print_function

import xml.etree.ElementTree as ET
import yagot


@yagot.garbage_checked()
def test_leaks_ET_Element_empty():
    """
    Test function with an empty Element object (i.e. no attributes,
    no child elements, no content).
    """
    _ = ET.Element('FOO')


@yagot.garbage_checked()
def test_leaks_ET_Element_one_attribute():
    """
    Test function with an Element object that has one attribute set.
    """
    elem = ET.Element('FOO')
    elem.attrib["BAR"] = 'bla'


@yagot.garbage_checked()
def test_leaks_ET_Element_one_child():
    """
    Test function with an Element object that has one child element.
    """
    elem = ET.Element('FOO')
    _ = ET.SubElement(elem, 'GOO')


@yagot.garbage_checked()
def test_leaks_ET_Element_one_child_text_attr():
    """
    Test function with an Element object that has one child element with
    text content and one attribute.
    """
    elem = ET.Element('FOO')
    subelem = ET.SubElement(elem, 'GOO')
    subelem.attrib["BAT"] = 'bla'
    subelem.text = 'some text'



# -*- coding: utf-8 -*-
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
The :func:`pywbem.siunit_obj` and :func:`pywbem.siunit` functions translate
the ``PUnit`` and ``Units`` qualifier values into human readable SI conformant
unit strings.

**Experimental:** *New in 1.1.0 as experimental*

Note: These functions do not perform any class operations; they take the
qualifiers as input.

The reason the ``Units`` qualifier is still supported is that the DMTF CIM
Schema (as of its version 2.49) still contains a number of schema elements that
have the ``Units`` qualifier but not the ``PUnit`` qualifier set.

The format and valid base units for the ``PUnit`` qualifier and the
valid values for the ``Units`` qualifier are defined in Annex C of
:term:`DSP0004`. Pywbem supports the definitions from :term:`DSP0004`
version 2.8, and the following additional ``Units`` qualifier values that are
used in DMTF CIM Schema version 2.49:

+--------------------------------------+
| Additional ``Units`` values          |
+======================================+
| ``-dBm``                             |
+--------------------------------------+
| ``Blocks``                           |
+--------------------------------------+
| ``Percentage``                       |
+--------------------------------------+
| ``Proportion``                       |
+--------------------------------------+
| ``Tenths of Revolutions per Minute`` |
+--------------------------------------+

By default, the string value returned from these functions may contain the
following Unicode characters outside of the 7-bit ASCII range. If the
``use_ascii`` parameter is `True`, these Unicode characters are replaced with
7-bit ASCII text as follows:

+---------------------------+---------+---------------+
| Unicode character code    | Unicode | 7-bit ASCII   |
+===========================+=========+===============+
| U+00B0: DEGREE SIGN       | ``°``   | ``deg``       |
+---------------------------+---------+---------------+
| U+00B5: MICRO SIGN        | ``µ``   | ``u``         |
+---------------------------+---------+---------------+
| U+2030: PER MILLE SIGN    | ``‰``   | ``1/1000``    |
+---------------------------+---------+---------------+
| U+2126: OHM SIGN          | ``Ω``   | ``Ohm``       |
+---------------------------+---------+---------------+
| U+00B2: SUPERSCRIPT TWO   | ``²``   | ``^2``        |
+---------------------------+---------+---------------+
| U+00B3: SUPERSCRIPT THREE | ``³``   | ``^3``        |
+---------------------------+---------+---------------+

Examples:

* ``PUnit("byte / second * 10^3")``-> ``kB/s``
* ``PUnit("byte * 2^10")``-> ``KiB``
* ``PUnit("hertz * 10^6")``-> ``MHz``
* ``PUnit("ampere * 10^-3")`` -> ``mA``
* ``Units("KiloBits per Second")`` -> ``kbit/s``
* ``Units("Tenths of Degrees C")`` -> ``1/10 °C``

Limitations:

* For PUnit qualifiers, vendor-defined base units are not supported
  (e.g. ``vendor:myunit``).

* For PUnit qualifiers, space characters within the parenthesis of
  ``decibel`` (e.g. ``decibel ( A )``) are not supported.

* For Units qualifiers, arbitrary numeric values that are part of the Units
  value (e.g. ``<numeric-value> NanoSeconds`` or
  ``Amps at <numeric-value> Volts``) are not generally supported, but
  only for those cases that are used in the DMTF CIM Schema (as of its version
  2.49):

  - ``250 NanoSeconds``
  - ``Amps at 120 Volts``
"""

import re
try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence
from nocasedict import NocaseDict
import six

from ._cim_obj import CIMProperty, CIMMethod, CIMParameter
from ._utils import _ensure_unicode

__all__ = ['siunit_obj', 'siunit']


def siunit_obj(cim_obj, use_ascii=False):
    # pylint: disable=line-too-long
    """
    Returns a human readable SI conformant unit string from the
    ``PUnit`` or ``Units`` qualifiers of the specified CIM object.

    **Experimental:** *New in 1.1.0 as experimental*

    If the CIM object has both the ``PUnit`` and ``Units`` qualifiers set, then
    ``PUnit`` is used and ``Units`` is ignored.

    Parameters:

      cim_obj (:class:`~pywbem.CIMProperty` or :class:`~pywbem.CIMMethod` or :class:`~pywbem.CIMParameter`):
        CIM object with qualifiers.

      use_ascii (:class:`py:bool`): Replace any Unicode characters in the
        returned string with 7-bit ASCII replacements,
        as described :mod:`above <pywbem._units>`.

    Returns:

      :term:`unicode string`: Human readable SI conformant unit string,
      or `None` if the CIM object has neither the ``PUnit`` nor the ``Units``
      qualifiers set.

    Raises:

      TypeError: Invalid type for cim_obj
      ValueError: Invalid format in PUnit qualifier
      ValueError: Unknown base unit in PUnit qualifier
      ValueError: Unknown unit in Units qualifier
    """  # noqa: E501
    # pylint: enable=line-too-long

    if not isinstance(cim_obj, (CIMProperty, CIMMethod, CIMParameter)):
        raise TypeError("Invalid type for cim_obj: {}, "
                        "must be CIMProperty, CIMMethod, or CIMParameter".
                        format(type(cim_obj)))

    punit_q = cim_obj.qualifiers.get('PUnit', None)
    punit = punit_q.value if punit_q else None

    units_q = cim_obj.qualifiers.get('Units', None)
    units = units_q.value if units_q else None

    return siunit(punit, units, use_ascii)


def siunit(punit=None, units=None, use_ascii=False):
    """
    Returns a human readable SI conformant unit string from the specified
    ``PUnit`` or ``Units`` qualifier values.

    **Experimental:** *New in 1.1.0 as experimental*

    If both ``punit`` and ``units`` are specified, then ``punit`` is used and
    ``units`` is ignored.

    Parameters:

      punit (:term:`string`):
        Value of the ``PUnit`` qualifier, or `None`.

      units (:term:`string`):
        Value of the ``Units`` qualifier, or `None`.

      use_ascii (:class:`py:bool`): Replace any Unicode characters in the
        returned string with 7-bit ASCII replacements,
        as described :mod:`above <pywbem._units>`.

    Returns:

      :term:`unicode string`: Human readable SI conformant unit string,
      or `None` if both qualifier value input parameters were `None`.

    Raises:

      TypeError: Invalid type for punit or unit
      ValueError: Invalid format in PUnit qualifier
      ValueError: Unknown base unit in PUnit qualifier
      ValueError: Unknown unit in Units qualifier
    """

    if punit is not None:
        punit = _ensure_unicode(punit)
        ret = _siunit_from_punit(punit)
    elif units is not None:
        units = _ensure_unicode(units)
        ret = _siunit_from_units(units)
    else:
        return None
    if use_ascii:
        ret = _uc2ascii(ret)
    return ret


# Unicode characters used in the output, by default
UC_DEGREE = u'\u00B0'  # U+00B0: DEGREE SIGN
# U+2103 (DEGREE CELSIUS) and U+2109 (DEGREE FAHRENHEIT) are intentionally
# not used, for better font/terminal compatibility.
UC_DEGREE_CELSIUS = u'{}C'.format(UC_DEGREE)  # Avoid U+2103: DEGREE CELSIUS
UC_DEGREE_FAHRENHEIT = u'{}F'.format(UC_DEGREE)  # Avoid U+2109: DEGREE FAHR.
UC_PERMILLE = u'\u2030'  # U+2030: PER MILLE SIGN
UC_MICRO = u'\u00B5'  # U+00B5: MICRO SIGN
UC_OHM = u'\u2126'  # U+2126: OHM SIGN
UC_SUP_2 = u'\u00B2'  # U+00B2: SUPERSCRIPT TWO
UC_SUP_3 = u'\u00B3'  # U+00B3: SUPERSCRIPT THREE
# U+2074 (SUPERSCRIPT FOUR) and adjacent characters up to U+2079
# (SUPERSCRIPT NINE) are intentionally not used, for better font/terminal
# compatibility.

# 7-bit ASCII replacements for the Unicode characters above
UC2ASCII = {
    UC_DEGREE: u'deg',
    UC_DEGREE_CELSIUS: u'degC',
    UC_DEGREE_FAHRENHEIT: u'degF',
    UC_PERMILLE: u'1/1000',
    UC_MICRO: u'u',
    UC_OHM: u'Ohm',
    UC_SUP_2: u'^2',
    UC_SUP_3: u'^3',
}

# Components in the PUnit regexp pattern, named consistent with ABNF in
# Annex C.1 of DSP0004
SIMPLE_NAME = r'[A-Za-z_][A-Za-z_\-0-9 ]*?'
WS = r'[ \t\n]?'  # optionality already included
SP = r'[ ]?'  # optionality already included
POSITIVE_WHOLE_NUMBER = r'[1-9][0-9]*'
POSITIVE_NUMBER = r'(?:{pwn}|(?:{pwn}|0)\.[0-9]*)'. \
    format(pwn=POSITIVE_WHOLE_NUMBER)
NUMBER = r'[\+\-]?{pn}'.format(pn=POSITIVE_NUMBER)
EXPONENT = r'[\+\-]?{pwn}'.format(pwn=POSITIVE_WHOLE_NUMBER)
DECIBEL_BASE_UNIT = r'decibel{sp}\({sn}\)'.format(sp=SP, sn=SIMPLE_NAME)

# The pattern for programmatic-unit.
# Because re returns only the last match of a group repetition, we need to
# parse the multiplied-base-unit and divided-base-unit components first as
# a whole, and then process them afterwards.
PUNIT = r'^({sn}|{dbu})' \
    r'((?:{ws}\*{ws}(?:{sn}|{dbu}))*)' \
    r'((?:{ws}/{ws}(?:{sn}|{dbu}))*)' \
    r'(?:{ws}(\*|/){ws}({n}))?' \
    r'(?:{ws}(\*|/){ws}({pwn}){ws}\^{ws}({e}))?$'. \
    format(sn=SIMPLE_NAME, dbu=DECIBEL_BASE_UNIT, ws=WS,
           pwn=POSITIVE_WHOLE_NUMBER, n=NUMBER, e=EXPONENT)
PUNIT_PATTERN = re.compile(PUNIT)

# The abbreviated SI units for each PUnit base unit defined in DSP0004
PUNIT_SIUNIT = NocaseDict([
    ("percent", "%"),
    ("permille", UC_PERMILLE),
    ("decibel", "dB"),
    ("count", ""),
    ("revolution", "rev"),
    ("degree", UC_DEGREE),
    ("radian", "rad"),
    ("steradian", "sr"),
    ("bit", "bit"),
    ("byte", "B"),
    ("dataword", "word"),
    ("MSU", "MSU"),
    ("meter", "m"),
    ("inch", "in"),
    ("rack unit", "U"),
    ("foot", "ft"),
    ("yard", "yd"),
    ("mile", "mi"),
    ("liter", "l"),
    ("fluid ounce", "fl.oz"),
    ("liquid gallon", "gal"),
    ("mole", "mol"),
    ("kilogram", "kg"),
    ("ounce", "oz"),
    ("pound", "lb"),
    ("second", "s"),
    ("minute", "min"),
    ("hour", "h"),
    ("day", "d"),
    ("week", "week"),
    ("gravity", "g"),
    ("degree celsius", UC_DEGREE_CELSIUS),
    ("degree fahrenheit", UC_DEGREE_FAHRENHEIT),
    ("kelvin", "K"),
    ("candela", "cd"),
    ("lumen", "lm"),
    ("nit", "nit"),
    ("lux", "lx"),
    ("newton", "N"),
    ("pascal", "Pa"),
    ("bar", "bar"),
    ("decibel(A)", "dB(A)"),
    ("decibel (A)", "dB(A)"),
    ("decibel(C)", "dB(C)"),
    ("decibel (C)", "dB(C)"),
    ("joule", "J"),
    ("watt", "W"),
    ("volt ampere", "VA"),
    ("volt ampere reactive", "var"),
    ("decibel(m)", "dBm"),
    ("decibel (m)", "dBm"),
    ("british thermal unit", "BTU"),
    ("ampere", "A"),
    ("coulomb", "C"),
    ("volt", "V"),
    ("farad", "F"),
    ("ohm", UC_OHM),
    ("siemens", "S"),
    ("weber", "Wb"),
    ("tesla", "T"),
    ("henry", "H"),
    ("becquerel", "Bq"),
    ("gray", "Gy"),
    ("sievert", "Sv"),
])

# The abbreviated SI prefixes for modifier1 in a PUnit.
# This dictionary covers only the typical values, assuming that for larger
# values, the modifier2 syntax is used.
# For modifier1 and modifier2, see the ABNF in Annex C.1 of DSP0004.
PUNIT_MOD1_SIPREFIX = {
    # mod1_num: si-prefix
    "1024": "Ki",
    "1000": "k",
    "100": "h",
    "10": "da",
    "1": "",
    "1/1": "",
    "1/10": "d",
    "1/100": "c",
    "1/1000": "m",
}

# The abbreviated SI prefixes for modifier2 in a PUnit.
# This dictionary covers all defined SI decimal prefixes and IEC binary
# prefixes.
# For modifier2, see the ABNF in Annex C.1 of DSP0004.
PUNIT_MOD2_SIPREFIX = {
    # (mod2_base, mod2_exp): si-prefix
    (2, 80): "Yi",
    (2, 70): "Zi",
    (2, 60): "Ei",
    (2, 50): "Pi",
    (2, 40): "Ti",
    (2, 30): "Gi",
    (2, 20): "Mi",
    (2, 10): "Ki",
    (10, 24): "Y",
    (10, 21): "Z",
    (10, 18): "E",
    (10, 15): "P",
    (10, 12): "T",
    (10, 9): "G",
    (10, 6): "M",
    (10, 3): "k",
    (10, 2): "h",
    (10, 1): "da",
    (10, -1): "d",
    (10, -2): "c",
    (10, -3): "m",
    (10, -6): "u",
    (10, -9): "n",
    (10, -12): "p",
    (10, -15): "f",
    (10, -18): "a",
    (10, -21): "z",
    (10, -24): "y",
}

# Dictionary to map unit power integer numbers to a suitable superscript
# Unicode character.
PUNIT_POWER = {
    1: u'',
    2: UC_SUP_2,
    3: UC_SUP_3,
}

# The abbreviated SI units for each Units unit defined in Annex C.2 of DSP0004.
UNITS_SIUNIT = NocaseDict([

    ("Bits", "bit"),
    ("KiloBits", "kbit"),
    ("MegaBits", "Mbit"),
    ("GigaBits", "Gbit"),

    ("Bits per Second", "bit/s"),
    ("KiloBits per Second", "kbit/s"),
    ("MegaBits per Second", "Mbit/s"),
    ("GigaBits per Second", "Gbit/s"),

    ("Bytes", "B"),
    ("KiloBytes", "kB"),  # Decimal prefix, according to SI
    ("MegaBytes", "MB"),  # Decimal prefix, according to SI
    ("GigaBytes", "GB"),  # Decimal prefix, according to SI
    ("Words", "words"),
    ("DoubleWords", "doublewords"),
    ("QuadWords", "quadwords"),

    ("Degrees C", UC_DEGREE_CELSIUS),
    ("Tenths of Degrees C", "1/10 " + UC_DEGREE_CELSIUS),
    ("Hundredths of Degrees C", "1/100 " + UC_DEGREE_CELSIUS),
    ("Degrees F", UC_DEGREE_FAHRENHEIT),
    ("Tenths of Degrees F", "1/10 " + UC_DEGREE_FAHRENHEIT),
    ("Hundredths of Degrees F", "1/100 " + UC_DEGREE_FAHRENHEIT),
    ("Degrees K", "K"),
    ("Tenths of Degrees K", "1/10 K"),
    ("Hundredths of Degrees K", "1/100 K"),
    ("Color Temperature", "K"),

    ("Volts", "V"),
    ("MilliVolts", "mV"),
    ("Tenths of MilliVolts", "1/10 mV"),
    ("Amps", "A"),
    ("MilliAmps", "mA"),
    ("Tenths of MilliAmps", "1/10 mA"),
    ("Watts", "W"),
    ("MilliWattHours", "mWh"),

    ("Joules", "J"),
    ("Coulombs", "C"),
    ("Newtons", "N"),

    ("Lumen", "lm"),
    ("Lux", "lx"),
    ("Candelas", "cd"),

    ("Pounds", "lb"),
    ("Pounds per Square Inch", "lb/in" + PUNIT_POWER[2]),

    ("Cycles", "cycles"),
    ("Revolutions", "rev"),
    ("Revolutions per Minute", "RPM"),
    ("Revolutions per Second", "rev/s"),

    ("Minutes", "min"),
    ("Seconds", "s"),
    ("Tenths of Seconds", "1/10 s"),
    ("Hundredths of Seconds", "1/100 s"),
    ("MicroSeconds", UC_MICRO + "s"),
    ("MilliSeconds", "ms"),
    ("NanoSeconds", "ns"),

    ("Hours", "h"),
    ("Days", "d"),
    ("Weeks", "week"),

    ("Hertz", "Hz"),
    ("MegaHertz", "MHz"),

    ("Pixels", "px"),
    ("Pixels per Inch", "px/in"),

    ("Counts per Inch", "counts/in"),

    ("Percent", "%"),
    ("Tenths of Percent", "1/10 %"),
    ("Hundredths of Percent", "1/100 %"),
    ("Thousandths", UC_PERMILLE),

    ("Meters", "m"),
    ("Centimeters", "cm"),
    ("Millimeters", "mm"),
    ("Cubic Meters", "m" + PUNIT_POWER[3]),
    ("Cubic Centimeters", "cm" + PUNIT_POWER[3]),
    ("Cubic Millimeters", "mm" + PUNIT_POWER[3]),

    ("Inches", "in"),
    ("Feet", "ft"),
    ("Cubic Inches", "in" + PUNIT_POWER[3]),
    ("Cubic Feet", "ft" + PUNIT_POWER[3]),
    ("Ounces", "oz"),
    ("Liters", "l"),
    ("Fluid Ounces", "fl.oz"),

    ("Radians", "rad"),
    ("Steradians", "sr"),
    ("Degrees", UC_DEGREE),

    ("Gravities", "g"),
    ("Pounds", "lb"),
    ("Foot-Pounds", "ft.lb"),

    ("Gauss", "G"),
    ("Gilberts", "Gb"),
    ("Henrys", "H"),
    ("MilliHenrys", "mH"),
    ("Farads", "F"),
    ("MilliFarads", "mF"),
    ("MicroFarads", UC_MICRO + "F"),
    ("PicoFarads", "pF"),

    ("Ohms", UC_OHM),
    ("Siemens", "S"),

    ("Moles", "mol"),
    ("Becquerels", "Bq"),
    ("Parts per Million", "ppm"),

    ("Decibels", "dB"),
    ("Tenths of Decibels", "1/10 dB"),

    ("Grays", "Gy"),
    ("Sieverts", "Sv"),

    ("MilliWatts", "mW"),

    ("DBm", "dBm"),

    ("Bytes per Second", "B/s"),
    ("KiloBytes per Second", "kB/s"),  # Decimal prefix, according to SI
    ("MegaBytes per Second", "MB/s"),  # Decimal prefix, according to SI
    ("GigaBytes per Second", "GB/s"),  # Decimal prefix, according to SI

    ("BTU per Hour", "BTU/h"),

    ("PCI clock cycles", "PCI clock cycles"),

    # The generic form of this unit (with arbitrary numeric value) is not
    # supported. The only usages in the DMTF CIM Schema (as of its version 2.49)
    # are the following:
    ("250 NanoSeconds", "250 ns"),

    ("Us", "U"),

    # The generic form of this unit (with arbitrary numeric value) is not
    # supported. The only usages in the DMTF CIM Schema (as of its version 2.49)
    # are the following:
    ("Amps at 120 Volts", "A (@120V)"),

    ("Clock Ticks", "clock ticks"),

    ("Packets", "packets"),
    ("per Thousand Packets", "/1000 packets"),

    # Additional values found in the DMTF CIM Schema (as of its version 2.49)
    # that are not defined as valid values in Annex C.2 of DSP0004 version 2.8
    ("-dBm", "-dBm"),
    ("Blocks", "blocks"),
    ("Percentage", "%"),
    ("Proportion", "proportion"),
    ("Tenths of Revolutions per Minute", "1/10 RPM"),
])


def _siunit_from_punit(punit):
    """
    Returns a human readable SI conformant unit string from the specified
    ``PUnit`` qualifier values.

    If the ``PUnit`` qualifier value does not conform to the syntax defined
    in DSP0004, ValueError is raised.

    Parameters:

      punit (:term:`string`): Value of ``PUnit`` qualifier. Must not be `None`.

    Returns:

      :term:`string`: Human readable SI conformant unit string.

    Raises:

      TypeError: Invalid type for punit
      ValueError: Unknown base unit in PUnit qualifier value
    """

    if not isinstance(punit, six.string_types):
        raise TypeError("Invalid type for punit: {}, must be string".
                        format(type(punit)))

    if punit == '':
        return ''

    # Parse the PUnit value
    m = PUNIT_PATTERN.match(punit)
    if m is None:
        raise ValueError(
            "Invalid format in PUnit qualifier: {!r}".format(punit))
    base_unit = m.group(1)
    mul_units_str = m.group(2)  # all multiplied-base-unit repetitions
    div_units_str = m.group(3)  # all divided-base-unit repetitions
    mod1_op = m.group(4)
    mod1_num = m.group(5)
    mod2_op = m.group(6)
    mod2_base = m.group(7)
    mod2_exp = m.group(8)

    # Process the units into mul_units and div_units
    _mul_units_str = base_unit
    if mul_units_str:
        _mul_units_str += mul_units_str
    mul_units = [s.strip() for s in _mul_units_str.split('*')]
    if div_units_str:
        div_units = [s.strip() for s in div_units_str.split('/')][1:]
    else:
        div_units = []

    if mod1_op:
        mod1_num = int(mod1_num)
    if mod2_op:
        mod2_base = int(mod2_base)
        mod2_exp = int(mod2_exp)

    # Translate the punits into SI units, combining repetitions into powers
    mul_siunit_str = _siunit_from_punit_base_units(mul_units)
    div_siunit_str = _siunit_from_punit_base_units(div_units)
    siunit_str = mul_siunit_str
    if div_siunit_str:
        siunit_str += '/' + div_siunit_str

    # Prefix for modifier2 (has precedence over modifier1)
    if mod2_op == '*':
        try:
            mod2_prefix = PUNIT_MOD2_SIPREFIX[(mod2_base, mod2_exp)]
        except KeyError:
            mod2_prefix = "{}^{} ".format(mod2_base, mod2_exp)
    elif mod2_op == '/':
        try:
            mod2_prefix = PUNIT_MOD2_SIPREFIX[(mod2_base, -mod2_exp)]
        except KeyError:
            mod2_prefix = "{}^{} ".format(mod2_base, -mod2_exp)
    else:
        mod2_prefix = ""

    # Prefix for modifier1
    if mod1_op == '*':
        mod1_str = "{}".format(mod1_num)
        if mod2_prefix:
            mod1_prefix = "{} ".format(mod1_str)
        else:
            try:
                mod1_prefix = PUNIT_MOD1_SIPREFIX[mod1_str]
            except KeyError:
                mod1_prefix = "{} ".format(mod1_str)
    elif mod1_op == '/':
        mod1_str = "1/{}".format(mod1_num)
        if mod2_prefix:
            mod1_prefix = "{} ".format(mod1_str)
        else:
            try:
                mod1_prefix = PUNIT_MOD1_SIPREFIX[mod1_str]
            except KeyError:
                mod1_prefix = "{} ".format(mod1_str)
    else:
        mod1_prefix = ""

    siunit_str = mod1_prefix + mod2_prefix + siunit_str

    return siunit_str


def _uc2ascii(uc_str):
    """
    Return the input string, where Unicode characters are replaced with their
    7-bit ASCII replacements as defined in UC2ASCII.
    """
    for uc in UC2ASCII:
        ac = UC2ASCII[uc]
        uc_str = uc_str.replace(uc, ac)
    return uc_str


def _siunit_from_punit_base_units(punit_base_units):
    """
    Return the (multiplied or divided) punit base units as an SI unit string.
    """
    assert isinstance(punit_base_units, Sequence)
    siunit_str = ''
    last_punit = None
    unit_power = 1
    for punit in punit_base_units:
        if punit == last_punit:
            unit_power += 1
        elif last_punit is not None:
            siunit_str += _get_punit_siunit(last_punit) + \
                _get_punit_power(unit_power)
            last_punit = None
            unit_power = 1
        last_punit = punit
    if last_punit is not None:
        siunit_str += _get_punit_siunit(last_punit) + \
            _get_punit_power(unit_power)
    return siunit_str


def _get_punit_siunit(punit):
    """
    Return the punit as an SI unit string, raising ValueError if it is unknown.
    """
    try:
        _siunit = PUNIT_SIUNIT[punit]
    except KeyError:
        new_exc = ValueError(
            "Unknown base unit in PUnit qualifier value: {}".format(punit))
        new_exc.__cause__ = None
        raise new_exc
    return _siunit


def _get_punit_power(power):
    """
    Return the power integer as a string.
    """
    assert isinstance(power, six.integer_types)
    try:
        _power = PUNIT_POWER[power]
    except KeyError:
        _power = "^{}".format(power)
    return _power


def _siunit_from_units(units):
    """
    Returns a human readable SI conformant unit string from the specified
    ``Units`` qualifier values.

    If the ``Units`` qualifier value is not one of the known units, ValueError
    is raised.

    Parameters:

      units (:term:`string`): Value of ``Units`` qualifier. Must not be `None`.

    Returns:

      :term:`string`: Human readable SI conformant unit string.

    Raises:

      TypeError: Invalid type for units
      ValueError: Unknown unit in Units qualifier
    """

    if not isinstance(units, six.string_types):
        raise TypeError("Invalid type for units: {}, must be string".
                        format(type(units)))

    try:
        return UNITS_SIUNIT[units]
    except KeyError:
        new_exc = ValueError(
            "Unknown unit in Units qualifier: {!r}".format(units))
        new_exc.__cause__ = None
        raise new_exc



"""
unichr2() utility function for pywbem testing.
"""

from __future__ import absolute_import, print_function

import sys
import six

SURROGATE_MIN_CP = 0xD800  # Lowest surrogate code point
SURROGATE_MAX_CP = 0xDFFF  # Highest surrogate code point
MAX_UNICODE_CP = 0x10FFFF  # Highest defined Unicode code point
MAX_UCS2_CP = 0xFFFF  # Highest Unicode code point in UCS-4


def unichr2(cp):
    """
    Like Python's built-in unichr(), just that it works for the entire
    Unicode character set in both narrow and wide Python builds, and with
    both Python 2 and Python 3.

    Returns None for surrogsate code points.
    """
    if SURROGATE_MIN_CP <= cp <= SURROGATE_MAX_CP:
        return None

    if sys.maxunicode == MAX_UNICODE_CP:
        # wide build
        return six.unichr(cp)

    if cp <= MAX_UCS2_CP:
        # narrow build, low range
        return six.unichr(cp)

    # narrow build, high range.
    # Python 2 narrow builds store unicode characters internally as
    # UTF-16. The unichr() function however only supports the low range.
    # Literals can be used to create characters in the high range (both
    # with high-range literals and with surrogates).
    # This implementation produces a unicode literal with surrogates
    # and evaluates that to c reate the unicode character.
    # Conversion of high range code points to UTF-16 surrogates:
    # Subtract 0x010000 from the code point, then take the high ten bits
    # of the result and add 0xD800 to get the first surrogate (in range
    # 0xD800-0xDBFF). Take the low ten bits of the result and add 0xDC00
    # to get the second surrogate (in range 0xDC00-0xDFFF).
    offset = cp - MAX_UCS2_CP - 1
    surr1 = (offset >> 10) + 0xD800
    surr2 = (offset & 0b1111111111) + 0xDC00
    lit = u'u"\\u%04X\\u%04X"' % (surr1, surr2)
    uchr = eval(lit)  # pylint: disable=eval-used
    return uchr



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
Internal module with utility stuff for command line programs.
"""

from __future__ import print_function, absolute_import

import argparse


class SmartFormatter(argparse.HelpFormatter):
    """Formatter class for `argparse`, that respects newlines in help strings.

    Idea and code from: https://stackoverflow.com/a/22157136

    Usage:
        If an argparse argument help text starts with 'R|', it will be treated
        as a *raw* string that does line formatting on its own by specifying
        newlines appropriately. The string should not exceed 55 characters per
        line. Indentation handling is still applied automatically and does not
        need to be specified within the string.

        Otherwise, the strings are formatted as normal and newlines are
        treated like blanks.

    Limitations:
        It seems this only works for the `help` argument of
        `ArgumentParser.add_argument()`, and not for group descriptions,
        and usage, description, and epilog of ArgumentParser.
    """

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        return argparse.HelpFormatter._split_lines(self, text, width)



#
# (C) Copyright 2020 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
The :class:`~pywbem_mock.InstanceWriteProvider` class provides the default
implementations of the ``CreateInstance``, ``ModifyInstance`` and
``DeleteInstance`` provider methods by means of
:meth:`~pywbem_mock.InstanceWriteProvider.CreateInstance`,
:meth:`~pywbem_mock.InstanceWriteProvider.ModifyInstance`, and
:meth:`~pywbem_mock.InstanceWriteProvider.DeleteInstance`.

A user-defined instance write provider may implement some or all of these
provider methods. The method descriptions linked above provide a detailed
definition of the input parameters, return values, and required behavior.

The following is a simple example of a user-defined instance write provider
that serves the 'CIM_Foo' and 'CIM_FooFoo' classes and implements
``CreateInstance`` for the purpose of setting a value for the 'InstanceID' key
property:

.. code-block:: python

    import uuid
    from pywbem_mock import InstanceWriteProvider

    class MyInstanceProvider(InstanceWriteProvider):

        # CIM classes served by this provider
        provider_classnames = ['CIM_Foo', 'CIM_FooFoo']

        def CreateInstance(self, namespace, new_instance):
            new_instance.properties["InstanceID"] = \\
                "{}.{}".format(new_instance.classname, uuid.uuid4())
            return super(MyInstanceProvider, self).CreateInstance(
                namespace, new_instance)

Because the provider in this example does not implement the
``ModifyInstance`` and ``DeleteInstance`` methods, these operations will
use their default implementations from
:class:`~pywbem_mock.InstanceWriteProvider`.

The provider in this example does not need to implement an ``__init__()``
method, because no additional init parameters are needed.
"""

from __future__ import absolute_import, print_function

from pywbem import CIMInstanceName, CIMError, \
    CIM_ERR_INVALID_PARAMETER, CIM_ERR_ALREADY_EXISTS

from pywbem._utils import _format

from ._baseprovider import BaseProvider

# None of the request method names conform since they are camel case
# pylint: disable=invalid-name

__all__ = ['InstanceWriteProvider']


class InstanceWriteProvider(BaseProvider):
    """
    This class defines those instance provider methods that may have user-
    defined providers that override the default provider implementation in this
    class.

    User providers are defined by creating a subclass of this class and
    defining a new provider method for one of the methods in this class with
    the same signature.

    Note that user-defined providers may, in turn, call the default providers
    in this class.
    """

    #: :term:`string`: Keyword defining the type of request the provider will
    #: service. The type for this provider class is predefined as 'instance'.
    provider_type = 'instance-write'

    def __init__(self, cimrepository=None):
        """
        Parameters:

          cimrepository (:class:`~pywbem_mock.BaseRepository` or subclass):
            The CIM repository to be used by the provider.
        """
        super(InstanceWriteProvider, self).__init__(cimrepository)

    def CreateInstance(self, namespace, new_instance):
        """
        Default provider method for
        :meth:`pywbem.WBEMConnection.CreateInstance`.

        Create a new CIM instance in the CIM repository of the mock WBEM server.

        Validation already performed by the provider dispatcher that calls
        this provider method:

        - The provider method is called only for the registered class and
          namespace (only applies to user-defined providers).

        - The Python types of all input parameters to this provider method are
          as specified below.

        - The namespace exists in the CIM repository.

        - The creation class of the new instance exists in the namespace
          in the CIM repository.

        - All properties specified in the new instance are exposed (i.e.
          defined and inherited with any overrides resolved) by the creation
          class in the CIM repository, and have the same type-related
          attributes (i.e. type, is_array, embedded_object).

        Validation that should be performed by this provider method:

        - new_instance does not specify any properties that are not
          allowed to be initialized by the client, depending on the model
          implemented by the provider.

        - new_instance specifies all key properties needed by the
          provider, depending on the model implemented by the provider.
          The CIM repository will reject any new instance that does not have
          all key properties specified.

        - The new instance (i.e. an instance with the new instance path) does
          not exist in the CIM repository.
          This validation needs to be done by the provider because the
          determination of the key properties for the new instance path may
          depend on the model implemented by the provider.
          The CIM repository will reject the creation of instances that already
          exist, so this check can be delegated to the repository once the
          new instance path has been determined.

        Parameters:

          namespace (:term:`string`):
            The name of the CIM namespace in which the CIM instance is to be
            created, in any lexical case, and with leading and trailing slash
            characters removed.

          new_instance (:class:`~pywbem.CIMInstance`):
            A representation of the CIM instance to be created.

            This object is a deep copy of the original client parameter, and may
            be modified by the provider as needed, before storing it in the
            CIM repository.

            The property names in this object have been adjusted to match the
            lexical case of the property definitions in the creation class
            of the instance in the CIM repository.

            The `classname` attribute of this object will specify the
            creation class for the new instance, in any lexical case.

            The `properties` attribute of this object will specify the
            initial property values for the new CIM instance, with property
            names in any lexical case. Key properties may or may not be
            included.

            The `path` attribute of this object will be `None`.

            The `qualifiers` attribute of this object, if non-empty, should
            be ignored by the provider, because instance-level qualifiers have
            been deprecated in CIM.

        Returns:

            :class:`~pywbem.CIMInstanceName`: Instance path of the new CIM
            instance.

        Raises:
            :exc:`~pywbem.CIMError`: The provider may raise CIMError with any
              status code, and typically raises:
              - CIM_ERR_INVALID_PARAMETER
              - CIM_ERR_ALREADY_EXISTS
        """

        # Get the creation class with all exposed properties and qualifiers.
        # Since the existence of the class has already been verified, this
        # will always succeed.
        class_store = self.cimrepository.get_class_store(namespace)
        creation_class = class_store.get(new_instance.classname, copy=False)

        # This default provider determines the instance path from the key
        # properties in the new instance. A user-defined provider may do that
        # as well, or invent key properties such as InstanceID.
        # Specifying strict=True in from_instance() verifies that all key
        # properties exposed by the class are specified in the new instance,
        # and raises ValueError if key properties are missing.
        try:
            new_instance.path = CIMInstanceName.from_instance(
                creation_class, new_instance, namespace=namespace, strict=True)
        except ValueError as exc:
            raise CIMError(CIM_ERR_INVALID_PARAMETER, str(exc))

        # Get the instance store of the CIM repository. Since the existence of
        # the namespace has already been verified, this will always succeed.
        instance_store = self.cimrepository.get_instance_store(namespace)

        # Store the new instance in the CIM repository, verifying that it does
        # not exist yet.
        try:
            instance_store.create(new_instance.path, new_instance)
        except ValueError:
            raise CIMError(
                CIM_ERR_ALREADY_EXISTS,
                _format("New instance {0!A} already exists in namespace "
                        "{1!A}.", new_instance.path, namespace))

        # CreateInstance returns the instance path of the new instance
        return new_instance.path

    def ModifyInstance(self, modified_instance, IncludeQualifiers=None):
        # pylint: disable=invalid-name,line-too-long,unused-argument
        """
        Default provider method for
        :meth:`pywbem.WBEMConnection.CreateInstance`.

        Modify an existing CIM instance in the CIM repository of the mock WBEM
        server.

        NOTE: This method specifies the namespace in modified_instance.path
        rather than as a separate input parameter.

        The modification of the instance in the CIM repository that is performed
        by the provider should be limited to property value changes (including
        addition of properties if not yet set on the instance), because
        instance-level qualifiers have been deprecated in CIM.

        The set of properties that are to be modified in the CIM instance has
        already been determined by the caller so that the modified_instance
        parameter specifies exactly the set of properties to be modified.
        Therefore, this provider method does not have a property list
        parameter.

        Validation already performed by the provider dispatcher that calls
        this provider method:

        - The provider method is called only for the registered class and
          namespace (only applies to user-defined providers).

        - The Python types of all input parameters to this provider method are
          as specified below.

        - The classnames in modified_instance are consistent between
          instance and instance path.

        - The namespace exists in the CIM repository.

        - The creation class of the instance to be modified exists in the
          namespace of the CIM repository.

        - The instance to be modified exists in the namespace of the CIM
          repository.

        - All properties in modified_instance that are to be modified are
          exposed (i.e. defined and inherited with any overrides resolved) by
          the creation class in the CIM repository, and have the same
          type-related attributes (i.e. type, is_array, embedded_object).

        - No key properties are requested to change their values.

        Validation that should be performed by this provider method:

        - modified_instance does not specify any changed values for
          properties that are not allowed to be changed by the client,
          depending on the model implemented by the provider.

        Parameters:

          modified_instance (:class:`~pywbem.CIMInstance`):
            A representation of the modified CIM instance, also indicating its
            instance path, with exactly the set of properties to be modified.

            This object is a deep copy of the original client parameter, and may
            be modified by the provider as needed, before storing it in the
            CIM repository.

            The `path` attribute of this object will be set and is the
            instance path of the instance to be modified in the CIM repository.
            Its `namespace`, `classname` and `keybindings` attributes
            will be set. The names will be in any lexical case.

            The `classname` attribute of this object will specify the creation
            class of the instance to be modified, in any lexical case.

            The `properties` attribute of this object will specify exactly the
            set of properties that are to be updated, taking into account the
            original ModifiedInstance and PropertyList input parameters of the
            ModifyInstance() client call.
            The lexical case of the property names has been adjusted to match
            the lexical cae of the property definitions in the creation class
            in the CIM repository.

            The `qualifiers` attribute of this object, if non-empty, should
            be ignored by the provider, because instance-level qualifiers have
            been deprecated in CIM.

          IncludeQualifiers (:class:`py:bool`):
            This parameter should be ignored by the provider, because
            instance-level qualifiers have been deprecated in CIM.

        Raises:
            :exc:`~pywbem.CIMError`: The provider may raise CIMError with any
              status code, and typically raises:
              - CIM_ERR_INVALID_PARAMETER
        """  # noqa: E501
        # pylint: disable=invalid-name,line-too-long

        namespace = modified_instance.path.namespace

        # Get a copy of the instance to be modified from the CIM repository.
        instance_store = self.cimrepository.get_instance_store(namespace)
        instance = instance_store.get(modified_instance.path)

        # Modify the properties of the local instance copy. The implemented
        # approach is intentionally careful, to ensure that only the property
        # values get updated.
        for pn in modified_instance.properties:
            mod_prop = modified_instance.properties[pn]
            inst_prop = instance.properties[pn]
            inst_prop.value = mod_prop.value  # Update the property value

        # Note that IncludeQualifiers is completely ignored.

        # Replace the instance in the CIM repository with the local copy.
        instance_store.update(modified_instance.path, instance)

    def DeleteInstance(self, InstanceName):
        """
        Default provider method for
        :meth:`pywbem.WBEMConnection.DeleteInstance`.

        Delete an existing instance in the CIM repository of the mock WBEM
        server.

        NOTE: This method specifies the namespace in InstanceName
        rather than as a separate input parameter.

        The provider is not responsible for determining other instances that
        depend on the instance to be deleted (e.g. association instances
        referencing the instance to be deleted); such dependency detection and
        handling is done elsewhere.

        Validation already performed by the provider dispatcher that calls
        this provider method:

        - The provider method is called only for the registered class and
          namespace (only applies to user-defined providers).

        - The Python types of all input parameters to this provider method are
          as specified below.

        - The namespace exists in the CIM repository.

        - The creation class of the instance to be deleted exists in the
          namespace of the CIM repository.

        - The instance to be deleted exists in the namespace of the CIM
          repository.

        Parameters:

          InstanceName (:class:`~pywbem.CIMInstanceName`):
            The instance path of the instance to be deleted with the
            following attributes:

            * `classname`: Name of the creation class of the instance.
            * `keybindings`: Keybindings of the instance.
            * `namespace`: Name of the CIM namespace containing the instance.
              Will not be `None`.
            * `host`: Will be `None`.

        Raises:

            None
        """

        namespace = InstanceName.namespace
        instance_store = self.cimrepository.get_instance_store(namespace)

        # Delete the instance from the CIM repository
        instance_store.delete(InstanceName)

    def post_register_setup(self, conn):
        """
        Method called by provider registration after registation of provider
        is successful. Using this method is optional for registration cases
        where the provider must execute some activity (ex. modify the
        CIM repository after successful provider registration).

        Override this method in the user-defined provider subclass to execute
        this method.

        Parameters:

          conn (:class:`~pywbem.WBEMConnection`):
            Current connection which allows client methods to be executed
            from within this method.
        """



#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
Test of pywbem_mock.ProviderRegistry.
"""

from __future__ import absolute_import, print_function

import pickle
import pytest

from ..utils.pytest_extensions import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import CIMClass  # noqa: E402
pywbem_mock = import_installed('pywbem_mock')
from pywbem_mock import FakedWBEMConnection, MethodProvider, \
    InstanceWriteProvider, BaseProvider  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


def assert_provreg_equal(provreg1, provreg2):
    """
    Assert that two ProviderRegistry objects are equal.

    Structure: self._registry[namespace][classname][prov_type] = prov_obj
    """
    ns_dict1 = provreg1._registry  # pylint: disable=protected-access
    ns_dict2 = provreg2._registry  # pylint: disable=protected-access

    assert sorted(ns_dict1) == sorted(ns_dict2)
    for ns in ns_dict1:
        class_dict1 = ns_dict1[ns]
        assert ns in ns_dict2
        class_dict2 = ns_dict2[ns]

        assert sorted(class_dict1.keys()) == sorted(class_dict2.keys())
        for cln in class_dict1.keys():
            ptype_dict1 = class_dict1[cln]
            assert cln in class_dict2
            ptype_dict2 = class_dict2[cln]

            assert sorted(ptype_dict1.keys()) == sorted(ptype_dict2.keys())
            for pt in ptype_dict1.keys():
                pobj1 = ptype_dict1[pt]
                assert pt in ptype_dict2
                pobj2 = ptype_dict2[pt]

                assert_prov_equal(pobj1, pobj2)


def assert_prov_equal(prov1, prov2):
    """
    Assert that two Provider objects (derived from BaseProvider) are equal.
    """
    assert isinstance(prov1, BaseProvider)
    assert isinstance(prov2, BaseProvider)
    assert prov1.__class__ == prov2.__class__
    assert prov1.provider_type == prov2.provider_type
    assert prov1.provider_classnames == prov2.provider_classnames
    # pylint: disable=protected-access
    assert prov1._interop_namespace_names == prov2._interop_namespace_names


def create_registered_conn(providers):
    """
    Create and return a FakedWBEMConnection object that has the specified
    providers registered.

    providers: Providers to be registered.
       providers are specified as tuple(prov_class, cim_class, namespace):
         - prov_class: Python class of provider
         - cim_class: Name of CIM class for the provider
         - namespace: CIM namespace for the provider
    """

    # Build a FakedWBEMConnection that has the required namespaces, classes
    # and registered providers in its provider registry.
    conn = FakedWBEMConnection()
    for item in providers:

        prov_class, cim_class, namespace = item

        # Make sure the namespace exists in the CIM repository
        if namespace not in conn.namespaces:
            conn.add_namespace(namespace)

        # Make sure the CIM class exists in the CIM repository
        class_store = conn.cimrepository.get_class_store(namespace)
        if not class_store.object_exists(cim_class):
            # An empty class is sufficient for this purpose:
            class_obj = CIMClass(cim_class)
            conn.add_cimobjects(class_obj, namespace=namespace)

        # Create the provider object, setting up its provider classes
        prov_class.provider_classnames = cim_class
        prov_obj = prov_class(conn.cimrepository)

        # Register the provider
        conn.register_provider(prov_obj, namespaces=namespace)

    return conn


class TstMethodProvider1(MethodProvider):
    """Test method provider #1"""

    provider_classnames = None  # Will be set before use

    def __init__(self, cimrepository):
        super(TstMethodProvider1, self).__init__(cimrepository)

    def InvokeMethod(self, methodname, localobject, params):
        """Dummy provider method"""
        pass


class TstMethodProvider2(MethodProvider):
    """Test method provider #2"""

    provider_classnames = None  # Will be set before use

    def __init__(self, cimrepository):
        super(TstMethodProvider2, self).__init__(cimrepository)

    def InvokeMethod(self, methodname, localobject, params):
        """Dummy provider method"""
        pass


class TstInstWrProvider1(InstanceWriteProvider):
    """Test instance write provider #1"""

    provider_classnames = None  # Will be set before use

    def __init__(self, cimrepository):
        super(TstInstWrProvider1, self).__init__(cimrepository)

    def CreateInstance(self, namespace, new_instance):
        """Dummy provider method"""
        pass


class TstInstWrProvider2(InstanceWriteProvider):
    """Test instance write provider #2"""

    provider_classnames = None  # Will be set before use

    def __init__(self, cimrepository):
        super(TstInstWrProvider2, self).__init__(cimrepository)

    def CreateInstance(self, namespace, new_instance):
        """Dummy provider method"""
        pass


# TODO: Add tests for __repr__()
# TODO: Add tests for display_registered_providers()
# TODO: Add tests for register_provider()
# TODO: Add tests for get_registered_provider()
# TODO: Add tests for provider_namespaces()
# TODO: Add tests for provider_classes()
# TODO: Add tests for provider_types()
# TODO: Add tests for provider_obj()


TESTCASES_PROVREG_ITERITEMS = [

    # Testcases for ProviderRegistry.iteritems()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * providers: providers for ProviderRegistry object
    #   providers are specified as tuple(prov_class, cim_class, namespace):
    #     - prov_class: Python class of provider
    #     - cim_class: Name of CIM class for the provider
    #     - namespace: CIM namespace for the provider
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty registry",
        dict(
            providers=[],
        ),
        None, None, True
    ),
    (
        "Registry with one provider",
        dict(
            providers=[
                (TstMethodProvider1, 'CIM_Foo1', 'root/blah1'),
            ],
        ),
        None, None, True
    ),
    (
        "Registry with prov for two different classes in same namespace",
        dict(
            providers=[
                (TstMethodProvider1, 'CIM_Foo1', 'root/blah1'),
                (TstInstWrProvider1, 'CIM_Foo1', 'root/blah1'),
                (TstMethodProvider2, 'CIM_Foo2', 'root/blah1'),
                (TstInstWrProvider2, 'CIM_Foo2', 'root/blah1'),
            ],
        ),
        None, None, True
    ),
    (
        "Registry with prov for same class in two different namespaces",
        dict(
            providers=[
                (TstMethodProvider1, 'CIM_Foo1', 'root/blah1'),
                (TstInstWrProvider1, 'CIM_Foo1', 'root/blah1'),
                (TstMethodProvider1, 'CIM_Foo1', 'root/blah2'),
                (TstInstWrProvider1, 'CIM_Foo1', 'root/blah2'),
            ],
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_PROVREG_ITERITEMS)
@simplified_test_function
def test_ProviderRegistry_iteritems(testcase, providers):
    """
    Test function for ProviderRegistry.iteritems().
    """
    conn = create_registered_conn(providers)
    provreg = conn._provider_registry  # pylint: disable=protected-access

    # The code to be tested
    res_providers = list(provreg.iteritems())

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Build lists for comparison, with list items that are:
    # tuple(namespace, cim_classname, prov_type, prov_classname)
    res_list = sorted([(p[0], p[1], p[2], p[3].__class__.__name__)
                       for p in res_providers])
    exp_list = sorted([(p[2], p[1], p[0].provider_type, p[0].__name__)
                       for p in providers])

    assert res_list == exp_list


TESTCASES_PROVREG_LOAD = [

    # Testcases for ProviderRegistry.load()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * providers1: providers for ProviderRegistry object #1
    #   * providers2: providers for ProviderRegistry object #2
    #   providers are specified as tuple(prov_class, cim_class, namespace):
    #     - prov_class: Python class of provider
    #     - cim_class: Name of CIM class for the provider
    #     - namespace: CIM namespace for the provider
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty registry loaded with empty registry",
        dict(
            providers1=[],
            providers2=[],
        ),
        None, None, True
    ),
    (
        "Empty registry loaded with registry with one provider",
        dict(
            providers1=[],
            providers2=[
                (TstMethodProvider1, 'CIM_Foo', 'root/blah'),
            ],
        ),
        None, None, True
    ),
    (
        "Registry with one provider loaded with empty registry",
        dict(
            providers1=[
                (TstMethodProvider1, 'CIM_Foo', 'root/blah'),
            ],
            providers2=[],
        ),
        None, None, True
    ),
    (
        "Registry with one provider loaded with registry with diff. provider",
        dict(
            providers1=[
                (TstMethodProvider1, 'CIM_Foo', 'root/blah'),
            ],
            providers2=[
                (TstInstWrProvider1, 'CIM_Foo', 'root/blah'),
            ],
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_PROVREG_LOAD)
@simplified_test_function
def test_ProviderRegistry_load(testcase, providers1, providers2):
    """
    Test function for ProviderRegistry.load().
    """
    conn1 = create_registered_conn(providers1)
    provreg1 = conn1._provider_registry  # pylint: disable=protected-access

    conn2 = create_registered_conn(providers2)
    provreg2 = conn2._provider_registry  # pylint: disable=protected-access

    # The code to be tested
    provreg1.load(provreg2)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert_provreg_equal(provreg1, provreg2)


TESTCASES_PROVREG_PICKLE = [

    # Testcases for pickling and unpickling ProviderRegistry objects

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * providers: providers for ProviderRegistry object to be used for test
    #   providers are specified as tuple(prov_class, cim_class, namespace):
    #     - prov_class: Python class of provider
    #     - cim_class: Name of CIM class for the provider
    #     - namespace: CIM namespace for the provider
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty provider registry",
        dict(
            providers=[],
        ),
        None, None, True
    ),
    (
        "Provider registry with one method provider",
        dict(
            providers=[
                (TstMethodProvider1, 'CIM_Foo', 'root/blah'),
            ],
        ),
        None, None, True
    ),
    (
        "Provider registry with one method provider",
        dict(
            providers=[
                (TstMethodProvider1, 'CIM_Foo', 'root/blah'),
                (TstInstWrProvider1, 'CIM_Foo', 'root/blah'),
            ],
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_PROVREG_PICKLE)
@simplified_test_function
def test_ProviderRegistry_pickle(testcase, providers):
    """
    Test function for pickling and unpickling ProviderRegistry objects.
    """
    conn = create_registered_conn(providers)
    provreg = conn._provider_registry  # pylint: disable=protected-access

    # Pickle the object
    pkl = pickle.dumps(provreg)

    # Unpickle the object
    provreg2 = pickle.loads(pkl)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert_provreg_equal(provreg2, provreg)



#!/usr/bin/env python

"""
Test CIM operations against a real WBEM server that is specified on the
command line.  It executes a specific test to test for correct operation of
the WBEMConnection timeout option in multiple modes of the server and client.
Specifically it tests both http and https schemes and confirms that the
timeout exception is received correctly over a wide range of timeout values.

Note that this test uses a specific class and method in OpenPegasus today to
execute an operation with a specified delay.

The default for this test takes several minutes to run because of size of
the maximum timeout (20 seconds)
"""

from __future__ import absolute_import

# pylint: disable=missing-docstring,superfluous-parens,no-self-use
import sys
import unittest
import warnings
import datetime
from getpass import getpass

import six

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from tests.utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import WBEMConnection, Uint32, ConnectionError, \
    TimeoutError  # noqa: E402 pylint: disable=redefined-builtin
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


# Identity of the OpenPegasus namespace, class, and method that implements
# the delayed response. NOTE: This is only available on OpenPegasus 2.15+.
NAMESPACE = 'test/testprovider'
TESTCLASS = 'Test_CLITestProviderClass'
TESTMETHOD = 'delayedMethodResponse'


class ElapsedTimer(object):
    """
        Set up elapsed time timer. Calculates time between initiation
        and access.
    """

    def __init__(self):
        """ Initiate the object with current time"""
        self.start_time = datetime.datetime.now()

    def reset(self):
        """ Reset the start time for the timer"""
        self.start_time = datetime.datetime.now()

    def elapsed_ms(self):
        """ Get the elapsed time in milliseconds. returns floating
            point representation of elapsed time in seconds.
        """
        dt = datetime.datetime.now() - self.start_time
        return ((dt.days * 24 * 3600) + dt.seconds) * 1000 + \
            dt.microseconds / 1000.0

    def elapsed_sec(self):
        """ get the elapsed time in seconds. Returns floating
            point representation of time in seconds
        """
        return self.elapsed_ms() / 1000


class ClientTest(unittest.TestCase):
    """ Base class that creates a pywbem.WBEMConnection for
        subclasses to use based on cmd line arguments. It provides
        a common cimcall method that executes test calls and logs
        results
    """

    def setUp(self):
        """Create a connection."""

        # pylint: disable=global-variable-not-assigned,global-statement
        global args                 # pylint: disable=invalid-name

        self.host = args['host']
        self.verbose = args['verbose']
        self.debug = args['debug']
        self.maxtimeout = args['maxtimeout']
        self.mintimeout = 2
        self.stop_on_err = args['stopOnErr']

        # set this because python 3 http libs generate many ResourceWarnings
        # and unittest enables these warnings.
        if not six.PY2:
            # pylint: disable=undefined-variable
            warnings.simplefilter("ignore", ResourceWarning)  # noqa: F821

    def connect(self, url, timeout=None):

        self.log('setup connection {0} timeout {1}'.format(url, timeout))
        conn = WBEMConnection(
            url,
            (args['username'], args['password']),
            NAMESPACE,
            timeout=timeout,
            no_verification=True,
            ca_certs=None)

        # enable saving of xml for display
        conn.debug = args['debug']
        self.log('Connected {0}'.format(url))
        return conn

    def log(self, data_):
        """Display log entry if verbose"""
        if self.debug:
            print('{0}'.format(data_))


class ServerTimeoutTest(ClientTest):

    def execute(self, url_type, timeout, delay):
        """
            Execute a single connect with timeout, followed by invokemethod
            of the method that delays response and test for possible responses.

            Tests for correct response type (good response if the delay is
            less than the timeout and timeout exception if the delay is greater
            than the timeout.  It ignores results where the operation delay
            and timeout are equal since the result could be either timeout
            or good response.
        """
        url = url_type + '://' + self.host
        conn = self.connect(url, timeout=timeout)
        execution_time = None

        if self.debug:
            print('execute url_type=%s timeout=%s delay=%s' % (url_type,
                                                               timeout, delay))

        err_flag = ""
        opttimer = ElapsedTimer()
        try:
            # Confirm server working with a simple request.
            conn.GetClass('CIM_ManagedElement')
            request_result = 0     # good response
            conn.InvokeMethod(TESTMETHOD, TESTCLASS,
                              [('delayInSeconds', Uint32(delay))])

            execution_time = opttimer.elapsed_sec()

            if delay > timeout:
                err_flag = "Timeout Error: delay gt timeout"
                if self.stop_on_err:
                    self.fail('should not get good response for %delay > %s' %
                              (delay, timeout))
            if delay == timeout:
                err_flag = "Good when timeout matches delay"

        # This exception terminates the test
        except ConnectionError as ce:
            request_result = 2    # Connection error received
            err_flag = "ConnectionError"
            execution_time = opttimer.elapsed_sec()

            self.fail('%s exception %s' % ('Failed ConnectionError)', ce))

        # this exception tests against expected result. It terminates the
        # test only if the timeout is not expected
        except TimeoutError:
            execution_time = opttimer.elapsed_sec()
            err_flag = "TimeoutError"
            execution_time = opttimer.elapsed_sec()
            request_result = 1    # timeout error received
            # error if the operation delay is lt timeout value and we get
            # a timeout
            if delay < timeout:
                err_flag = 'Error. delay < timeout'
                if self.stop_on_err:
                    self.fail('should not get timeout for %delay < %s' %
                              (delay, timeout))
            if delay == timeout:
                err_flag = "Timeout when timeout matches delay"

        # This exception terminates the test if stop_on_err set.
        except Exception as ec:      # pylint: disable=broad-except
            err_flag = "Test Failed.General Exception"
            execution_time = opttimer.elapsed_sec()
            request_result = 2
            if self.stop_on_err:
                self.fail('%s exception %s' % ('Failed(Exception)', ec))

        # generate table entry if verbose. This defines result for this
        # test
        if self.verbose:
            request_result_txt = ['Good Rtn', 'Timeout ', 'Failure']

            print('%-5s %-7s %7d %5d %10.2f  %s' % (
                url_type,
                request_result_txt[request_result],
                timeout, delay,
                execution_time,
                err_flag))

        return (request_result == 1)

    def test_all(self):
        """ Tests all of the variations in a loop."""

        print('url   result   timeout delay  time(sec)  Comments')

        for url_type in ['https', 'http']:
            for timeout in range(self.mintimeout, self.maxtimeout):

                # loop while good response received to test range of
                # server delays. Test for range around current timeout value
                break_loop = False
                min_delay = timeout - 2
                max_delay = timeout * 3
                good_rtn_rcvd = False
                timeout_rcvd = False
                for delay in range(min_delay, max_delay):
                    timed_out = self.execute(url_type, timeout, delay)
                    if not timed_out:
                        good_rtn_rcvd = True
                    if timed_out:
                        timeout_rcvd = True

                    # After return indicates timeout, run one more test
                    # then break delay loop. Assures that the timeour
                    # response is consistent.
                    if break_loop:
                        break
                    if timed_out:
                        break_loop = True
                # confirm that both a good response and timeout received for
                # this delay range.
                self.assertTrue(timeout_rcvd, "Timeout never received")
                self.assertTrue(good_rtn_rcvd, "Good return never received")


def parse_args(argv_):
    argv = list(argv_)

    if len(argv) <= 1:
        print('Error: No arguments specified; Call with --help or -h for '
              'usage.')
        sys.exit(2)
    elif argv[1] == '--help' or argv[1] == '-h':
        print('')
        print('Test program for client timeout of cim operations.\n')
        print('Requires OpenPegasus because it uses a special method\n')
        print('that delays the response for a defined time.\n')
        print('Tests a range of timeout values against http\n ')
        print('and https with operation that generates known delay.\n')
        print('Complete test is in a single unittest function')
        print('')
        print('Usage:')
        print('    %s [GEN_OPTS] URL [USERNAME%%PASSWORD [UT_OPTS '
              '[UT_CLASS ...]]] ' % argv[0])
        print('')
        print('Where:')
        print('    GEN_OPTS            General options (see below).')
        print('    HOST                Name of the target WBEM server.\n'
              '                        No Scheme prefix\n'
              '                        defines ssl usage')
        print('    USERNAME            Userid used to log into\n'
              '                        WBEM server.\n'
              '                        Requests user input if not supplied')
        print('    PASSWORD            Password used to log into\n'
              '                        WBEM server.\n'
              '                        Requests user input if not supplier')
        print('')
        print('General options[GEN_OPTS]:')
        print('    --help, -h          Display this help text.')
        print('    -t MAXTIMEOUT       Maximum timout value in sec. to test.\n'
              '                        default is 20 sec.')
        print('    -v                  Verbose output which includes\n'
              '                        result of each test.')
        print('    -s                  Stop test on first timeout error.\n'
              '                        Otherwise errors in timeout are just\n'
              '                        reported. Other errors stop test.')
        print('    -d                  Debug flag for extra displays')

        print('------------------------')
        print('Unittest arguments[UT_OPTS]:')
        print('')
        sys.argv[1:] = ['--help']
        unittest.main()
        sys.exit(2)

    args_ = {}
    # set argument defaults
    args_['maxtimeout'] = 20
    args_['stopOnErr'] = False
    args_['verbose'] = False
    args_['username'] = None
    args_['password'] = None
    args_['debug'] = False

    # options must proceed arguments
    while True:
        if argv[1][0] != '-':
            # Stop at first non-option
            break

        if argv[1] == '-t':
            args_['maxtimeout'] = int(argv[2])
            del argv[1:3]
        elif argv[1] == '-d':
            args_['delay'] = int(argv[2])
            del argv[1:3]
        elif argv[1] == '-v':
            args_['verbose'] = True
            del argv[1:2]
        elif argv[1] == '-d':
            args_['debug'] = True
            del argv[1:2]
        elif argv[1] == '-s':
            args_['stopOnErr'] = True
            del argv[1:2]
        elif argv[1] == '-d':
            args_['debug'] = True
            del argv[1:2]
        else:
            print("Error: Unknown option: %s" % argv[1])
            sys.exit(1)

    args_['host'] = argv[1]
    del argv[1:2]

    if len(argv) >= 2:
        args_['username'], args_['password'] = argv[1].split('%')
        del argv[1:2]
    else:
        # Get user name and pw from console
        sys.stdout.write('Username: ')
        sys.stdout.flush()
        args_['username'] = sys.stdin.readline().strip()
        args_['password'] = getpass()
    return args_, argv


if __name__ == '__main__':
    args, sys.argv = parse_args(sys.argv)  # pylint: disable=invalid-name

    if args['verbose'] in args:
        print("Using WBEM Server:")
        print("  host: %s" % args['host'])
        print("  username: %s" % args['username'])
        if args['password'] is not None:
            print("  password: %s" % ("*" * len(args['password'])))
        print("  maxtimeout: %s" % args['maxtimeout'])
        print("  verbose: %s" % args['verbose'])
        print("  stopOnErr: %s" % args['stopOnErr'])
        print("  debug: %s" % args['debug'])

    # Note: unittest options are defined in separate args after
    # the url argument.

    unittest.main()
